from reportfunctions import TextTable, header, location


def create(info, panel, identifier, outputstem):

    if "depth" not in panel:
        return

    with file(outputstem+"_amplicon_data.tsv", "wt") as f:
        template = "{0}\t{1}\t{2}\t{3}\t{4}\t{5}\n"
        f.write(template.format("identifier", "run", "amplicon", "chr", "fdepth", "rdepth"))
        for i in info:
            f.write(template.format(identifier.name, identifier.run, i.name, i.chrom, i.f_depth, i.r_depth))

    report = header(panel, identifier)

    passed = 0
    mean_depth = 0 
    for i in info:
        if i.minimum_depth >= panel.depth:
            passed += 1
            mean_depth += i.mean_depth     
    mean_depth = ", (mean depth {0})".format(mean_depth/passed) if (passed>0) else ""
    report += ["\n", "{0} of {1} amplicons ({2:0.1f}%) provided adequate coverage{3}\n".format(passed, len(info), float(passed)/len(info)*100, mean_depth)]
  
    table = TextTable()
    table.headers.append(["Amplicon", "Gene", "Location", "Forward", "Reverse"])
    table.headers.append(["", "", "", "Depth", "Depth"])
    for i in info:
        if i.minimum_depth < panel.depth:
            table.rows.append([i.name, location(i.gr, panel), i.gr.locations_as_string, i.f_depth, i.r_depth])
    if len(table.rows) > 0:
        report += ["\n\n", "The following amplicons failed to provide adequate coverage\n"] + table.formated(sep="   ")
 
    table = TextTable()
    table.headers.append(["Amplicon", "Gene", "Location", "Forward", "Reverse", "Ratio"])
    table.headers.append(["", "", "", "Depth", "Depth", ""])
    for i in info:
        if  i.ratio < 0.7 and i.maximum_depth > panel.depth:
            table.rows.append([i.name, location(i.gr, panel), i.gr.locations_as_string, i.f_depth, i.r_depth, (i.ratio, "{:.2f}")])
    if len(table.rows) > 0:
        report += ["\n\n", "The following amplicons had an abnormal forward:reverse depth ratio (<0.7)\n"] + table.formated(sep="   ")
    
    with file(outputstem+"_covermi_technical_report.txt", "wt") as f:
        f.writelines(report)




import sys, os, tkFileDialog, Tkinter, pdb


def main():
    rootwindow = Tkinter.Tk()
    rootwindow.withdraw()

    print "BringMiUp"
    print "Please select a folder"
    folderpath = tkFileDialog.askdirectory(parent=rootwindow, title='BringMiUp. Please select a folder')
    if folderpath == "":
        sys.exit()

    skiplist = []
    print "{0} folder selected".format(os.path.basename(folderpath))
    for root, dirnames, filenames in os.walk(folderpath):
        if root != folderpath:
            for filename in filenames:
                if filename.endswith(".bam") or filename.endswith(".bai") or filename.endswith(".vcf"):

                    oldfile = os.path.join(root, filename)
                    newfile = os.path.join(folderpath, filename)
                    if os.path.exists(newfile):
                        skiplist.append(oldfile)
                        print "File \""+oldfile+"\" already exists in \""+folderpath+"\". Skipping"
                    else:
                        print "Moving "+filename
                        os.rename(oldfile, newfile)

    if len(skiplist) > 0:
        print "WARNING. The following files were not moved"
        print ", ".join(skiplist)


if __name__ == "__main__":
    main()



#!/usr/bin env python
import sys, os, time, re, getopt, pdb
from collections import namedtuple
from cov import Cov, fake_paired_end_reads
from panel import Panel
from version import __version__
import covermiconf, technicalreport, clinicalreport, designreport, covermiplot
from covermiexception import CoverMiException
#import testreport as clinicalreport



illumina_suffix = re.compile("(.+)_S[0-9]+$")
Sample = namedtuple("Sample", "name run path")

def walk_samples(path, ext=".bam", strip_trailing_underscore=True):
    found = []

    if os.path.isfile(path):
        if not os.path.splitext(path)[1] == ext:
            raise CoverMiException("file "+path+" is not of type "+ext)
        path_noext, extension = os.path.splitext(path)
        runpath, sample = os.path.split(path_noext)
        run = os.path.basename(runpath)
        yield Sample(strip_underscore(sample), run, path)

    elif os.path.isdir(path):
        for fn in os.listdir(path):
            if os.path.isfile(os.path.join(path, fn)) and os.path.splitext(fn)[1] in ext: # Single run format
                for sample in walk_run(path, ext):
                    yield sample
                return

        found = [] # Muliptle run format
        for fn in os.listdir(path):
            if os.path.isdir(os.path.join(path, fn)):
               for sample in walk_run(os.path.join(path, fn), ext):
                    yield sample

    else:
        raise CoverMiException(path+" is not a file or directory")



def walk_run(path, ext):
    run = os.path.basename(path)
    for root, dirnames, filenames in os.walk(path):
        for fn in filenames:
            sample, extension = os.path.splitext(fn)
            if os.path.isfile(os.path.join(root, fn)) and extension == ext:
                yield Sample(strip_underscore(sample), run, os.path.join(root, fn))
        


def strip_underscore(sample):
    matchobj = illumina_suffix.search(sample)
    return matchobj.group(1) if matchobj else sample 



def create_output_dir(output_path, bam_path):
    output_path = os.path.join(output_path, "{0}_covermi_output".format(os.path.splitext(os.path.basename(bam_path))[0]))
    try:
        os.mkdir(output_path)
    except OSError:
        if os.path.isdir(output_path):
            raise CoverMiException("{0} folder already exists".format(output_path))
        else:
            raise CoverMiException("Unable to create folder {0}".format(output_path))
    return output_path


def main(panel_path="", bam_path="", output_path=""):

    print "CoverMi v{}".format(__version__)

    try:
        opts, args = getopt.getopt(sys.argv[1:], "", ["args", "depth"])
    except getopt.GetoptError as err:
        raise CoverMiException(str(err))

    if len(args) == 3:
        bam_path, output_path, panel_path = args

    depth = None
    for opt, val in opts:
        if opt == "--args":
            print "Arguments = "+" ".join(["\"{}\"".format(a) for a in (bam_path, output_path, panel_path)])
        elif opt == "--depth":
            depth = int(val)

    if not os.path.exists(panel_path):
        panel_path = os.path.join(covermiconf.load_conf()["panel_root"], panel_path)

    panel = Panel(panel_path, verbose=True)
    if depth is not None:
        panel.options["Depth"] = int(depth)
    output_path = create_output_dir(output_path, bam_path if bam_path!="" else panel_path)
    print "Processing..."

    if bam_path != "":
        bam_file_list = list(sample for sample in walk_samples(bam_path))
        if len(bam_file_list) == 1:
            clinical_report_path = output_path
            technical_report_path = output_path
        else:
            clinical_report_path = os.path.join(output_path, "clinical")
            technical_report_path = os.path.join(output_path, "technical")
            os.mkdir(clinical_report_path)
            os.mkdir(technical_report_path)

        output_stems = set([])
        for sample in bam_file_list:
            start_time = time.time()
            print "{0}/{1}".format(sample.run, sample.name)

            output_stem = sample.name
            dup_num = 1
            while output_stem in output_stems:
                output_stem = "{0}({1})".format(sample.name, dup_num)
                dup_num += 1
            output_stems.add(output_stem)

            extra = {"amplicons": panel.amplicons} if "amplicons" in panel else {"locations": panel.transcripts, "print_progress": True}
            cov = Cov(sample.path, **extra)

            if cov.amplicon_info: technicalreport.create(cov.amplicon_info, panel, sample, os.path.join(technical_report_path, output_stem))
            clinicalreport.create(cov, panel, sample, os.path.join(clinical_report_path, output_stem))
            covermiplot.plot(cov, panel, sample.name, os.path.join(clinical_report_path, output_stem+"_plot.pdf"))

            seconds = int(time.time() - start_time)
            time_string = "{0} sec".format(seconds) if (seconds<60) else "{0} min {01} sec".format(seconds/60, seconds%60)
            print"file {0} of {1} completed in {2}".format(len(output_stems), len(bam_file_list), time_string)
    else:
        panel_name = panel.file_names["Panel"]
        cov = Cov(fake_paired_end_reads(panel.amplicons, depth=panel.depth*2))
        designreport.create(cov, panel, None, os.path.join(output_path, panel_name))
        covermiplot.plot(cov, panel, panel_name, os.path.join(output_path, panel_name+"_plot.pdf"))

if __name__ == "__main__":
    main()




# panel is a dict containing all of the data structures that define a panel
# 	Amplicons:		genomic range
# 	Exons:			genomic range
# 	Transcripts:		genomic range
# 	Depth:
#	Variants_Disease:	genomic range
#	Variants_Gene:		genomic range
#	Variants_Mutation:	genomic range
#	Filenames:		dict of all files in the panel directory with filetype as the key
#       Options:		dict of all options, including depth

# the following are only included for a design panel
#	 AllTranscripts: 	genomic range
#	 AllExons:		genomic range
#	 Excluded:		list of excluded amplicons

import os, re, pdb
from gr import Gr
from reader import Bed, Vcf, IlluminaManifest, RefFlat, Variants
from covermiexception import CoverMiException


regexps = (("Reference",        re.compile(".+?\t.+?\tchr.+?\t[+-]\t[0-9]+\t[0-9]+\t[0-9]+\t[0-9]+\t[0-9]+\t[0-9,]+\t[0-9,]+\\s*$")), #Eleven columns - refflat
           ("Depth",            re.compile("[0-9]+\\s*$")), #Single number - depth
           ("Targets",          re.compile("[a-zA-Z][a-zA-Z0-9-]*\\s*$")), #Single column - targets
           ("Targets",          re.compile("[a-zA-Z][a-zA-Z0-9-]* +[a-zA-Z0-9_]+\\s*$")), #Single column - targets
           ("Manifest",         re.compile("\\[Header\\]$")), #Manifest
           ("Variants",         re.compile(".+?\t.+?\t.+?\t[a-zA-Z0-9]+\t(null|chr[1-9XYM][0-9]?)\t(null|[0-9]+)\t(null|[0-9]+)\t(null|[+-])\t")), #Nine columns - variants
           ("Canonical",        re.compile("[a-zA-Z][a-zA-Z0-9]*\t[a-zA-Z][a-zA-Z0-9-]* +[a-zA-Z0-9_]+\\s*$")), #Two columns - canonical
           ("Canonical",        re.compile("[a-zA-Z][a-zA-Z0-9]*\t[a-zA-Z0-9_]+\\s*$")), #Two columns - canonical
           ("DesignStudio",     re.compile("chr[0-9XYM]+\t[0-9]+\t[0-9]+[\t\n]")), #Six column Bedfile - design
           ("Disease_Names",    re.compile("#Variants Disease Name Translation\\s*$")), #Disease_Names
           ("SNPs",             re.compile("chr[1-9XYM][0-9]?:[0-9]+ [ATCG\.]+>[ATCG\.]\\s*$")), #SNPs
           ("VCF",              re.compile("##fileformat=VCFv[0-9\.]+\\s*$")), #VCF
           ("Options",          re.compile("#CoverMi options\\s*$")), # CoverMi panel options
          )



def Panels(panels_path, **kwargs):
    panels = {}
    for name in os.listdir(panels_path):
        panel_path = os.path.join(panels_path, name)
        if os.path.isdir(panel_path):
            try:
                panels[name] = Panel(panel_path, **kwargs)
            except CoverMiException:
                continue
    return panels



class Panel(object):

    @classmethod
    def build(cls, **kwargs):
        panel = cls()
        for attr, value in kwargs.items():
            setattr(panel, "_"+attr, value)
        return panel


    def __init__(self, panel_path=None, verbose=False, required=()):
        self._amplicons = None
        self._offtarget = None
        self._expected = None
        self._unexpected = None
        self._transcripts = None
        self._exons = None
        self._alltranscripts = None
        self._allexons = None
        self._variants_disease = None
        self._variants_mutation = None
        self._variants_gene = None
        self._options = None

        self.panel_path = panel_path
        self._verbose = verbose
        self._refflat_verbose = verbose
        self._variants_verbose = verbose

        self.file_names = {}

        if panel_path:
            self.file_names["Panel"] = os.path.basename(panel_path)
            found = set()
            for fn in os.listdir(panel_path):
                file_path = os.path.join(panel_path, fn)
                if not os.path.isfile(file_path): continue
                file_extension = os.path.splitext(fn)[1]
                if file_extension in (".bam", ".vcf", ".bai"): continue

                with open(file_path, "rU") as f:
                    testlines = [f.readline(), f.readline()]
                
                alreadyfound = ""
                for testline in testlines:
                    testline = testline.strip()
                    for filetype, regexp in regexps:
                        if regexp.match(testline):
                            if alreadyfound != "":
                                raise CoverMiException("Not a valid CoverMi panel\nUnable to uniquely identify file {0}, matches both {1} and {2} file formats".format(filepath, filetype, alreadyfound))
                            if filetype in self.file_names:
                                raise CoverMiException("Not a valid CoverMi panel\nFiles {} and {} both match {} format".format(filepath, self.file_names[filetype], filetype))
                            alreadyfound = filetype
                            self.file_names[filetype] = fn
                    if alreadyfound:
                        break

        for attr in required:
            test = getattr(self, attr)


    @property
    def amplicons(self):
        if self._amplicons is None:
            if "Manifest" in self.file_names and "DesignStudio" in self.file_names:
                raise CoverMiException("Both Manifest and Design Studio Files found in {0}".format(self.panel_path))
            if "Manifest" not in self.file_names and "DesignStudio" not in self.file_names:
                raise CoverMiException("Unable to identify a Manifest file or a Design Studio Bedfile in {0}".format(self.panel_path))
            if "Manifest" in self.file_names:
                if self._verbose: print "Loading manifest file: {0}".format(self.file_names["Manifest"])
                manifest = IlluminaManifest(os.path.join(self.panel_path, self.file_names["Manifest"]))
                self._amplicons = Gr(manifest)
                self._offtarget = Gr(manifest.offtarget())
            elif "DesignStudio" in self.file_names:
                if self._verbose: print "Loading Design Studio amplicons bedfile: {0}".format(self.file_names["DesignStudio"])
                self._amplicons =  Gr(Bed(os.path.join(self.panel_path, self.file_names["DesignStudio"])))
        return self._amplicons


    @property
    def offtarget(self):
        if self._offtarget is None:
            amplicons = self.amplicons
        return self._offtarget


    @property
    def expected(self):
        if self._expected is None:
            self._expected = self.amplicons.combined_with(self.offtarget).merged
        return self._expected


    @property
    def unexpected(self):
        if self._unexpected is None:
            self._unexpected = self.expected.inverted
        return self._unexpected


    def common_ref_flat_checks(self):
        extra = {}
        if "Reference" not in self.file_names:
            raise CoverMiException("Unable to identify a Reference file in {0}".format(self.panel_path))
        if self._refflat_verbose: print "Loading reference file: {0}".format(self.file_names["Reference"])

        if "Canonical" in self.file_names:
            extra["canonical"] = open(os.path.join(self.panel_path, self.file_names["Canonical"]), "rb")
            if self._refflat_verbose: print "Loading canonical gene list: {0}".format(self.file_names["Canonical"])
        else:
            if self._refflat_verbose: print "WARNING. No canonical gene list found. Loading all transcripts of targeted genes"        
        self._refflat_verbose = False
        return extra


    @property
    def transcripts(self):
        if self._transcripts is None:
            extra = self.common_ref_flat_checks()

            if "Targets" in self.file_names:
                extra["genes"] = open(os.path.join(self.panel_path, self.file_names["Targets"]), "rb")
                if self._verbose: print "Loading targeted genes file: {0}".format(self.file_names["Targets"])
                if "amplicons" not in self:
                    if self._verbose: print "No amplicons in panel. Will perform coverage analysis over exons of genes in targets file"

            refflatreader = RefFlat(os.path.join(self.panel_path, self.file_names["Reference"]), **extra)
            self._exons = Gr(refflatreader.read(RefFlat.EXONS))                                                     
            self._transcripts = Gr(refflatreader.read(RefFlat.TRANSCRIPTS))

            if "Targets" not in self.file_names:
                if self._verbose: print "WARNING. No file identifying targeted genes. Loading all genes touched by an amplicon"
                self._transcripts = self._transcripts.touched_by(self.amplicons) # Will raise an uncaught exception if not present, as intended
                self._exons = self._exons.touched_by(self._transcripts).subset(self._transcripts.names)
        return self._transcripts
 

    @property
    def exons(self):
        if self._exons is None:
            transcripts = self.transcripts
        return self._exons


    @property
    def alltranscripts(self):
        if self._alltranscripts is None:
            extra = self.common_ref_flat_checks()

            refflatreader = RefFlat(os.path.join(self.panel_path, self.file_names["Reference"]), **extra)
            self._allexons = Gr(refflatreader.read(RefFlat.EXONS))                                                     
            self._alltranscripts = Gr(refflatreader.read(RefFlat.TRANSCRIPTS)) 
        return self._alltranscripts
 

    @property
    def allexons(self):
        if self._allexons is None:
            alltranscripts = self.alltranscripts
        return self._allexons


    def common_variants_checks(self):
        extra = {}
        if self._variants_disease is None:
            if "Variants" not in self.file_names:
                raise CoverMiException("Unable to identify a Variants file in {0}".format(self.panel_path))
        if self._variants_verbose: print "Loading variants file: {0}".format(self.file_names["Variants"])
        if "Disease_Names" in self.file_names:
            if self._variants_verbose: print "Loading disease names file: {0}".format(self.file_names["Disease_Names"])
            extra["disease_names"] = open(os.path.join(self.panel_path, self.file_names["Disease_Names"]), "rb")
        self._variants_verbose = False    
        return extra


    @property
    def variants_disease(self):
        if self._variants_disease is None:
            extra = self.common_variants_checks()
            self._variants_disease = Gr(Variants(os.path.join(self.panel_path, self.file_names["Variants"]), **extra).read(Variants.DISEASE))        
        return self._variants_disease


    @property
    def variants_gene(self):
        if self._variants_gene is None:
            extra = self.common_variants_checks()
            self._variants_gene = Gr(Variants(os.path.join(self.panel_path, self.file_names["Variants"]), genes=self.transcripts.names, **extra).read(Variants.GENE))

            # Check to try and see if the variants file is aligned against the correct reference genome
            targeted_variants = self._variants_gene.subset(self.transcripts.names)
            variants_in_correct_location = targeted_variants.touched_by(self.transcripts).number_of_components * 100 // targeted_variants.number_of_components
            if variants_in_correct_location < 98:
                if self._verbose: print "WARNING. Only {}% of targeted variants are within targeted genes. ?Correct reference genome".format(variants_in_correct_location)
        return self._variants_gene


    @property
    def variants_mutation(self):
        if self._variants_mutation is None:
            extra = self.common_variants_checks()
            self._variants_mutation = Gr(Variants(os.path.join(self.panel_path, self.file_names["Variants"]), genes=self.transcripts.names, **extra).read(Variants.MUTATION))
        return self._variants_mutation


    @property
    def depth(self):
        try:
            return self.options["Depth"]
        except KeyError:
            raise CoverMiException("Depth not available")


    @property
    def options(self): 
        if self._options is None:
            self._options = {}
            if "Depth" in self.file_names:
                with open(os.path.join(self.panel_path, self.file_names["Depth"]), "rU") as f:
                    self._options["Depth"] = f.read().strip()
            if "Options" in self.file_names:
                if self._verbose: print "Loading options file: {0}".format(self.file_names["Options"])
                with open(os.path.join(self.panel_path, self.file_names["Options"]), "rU") as f:
                    for line in f:
                        line = line.strip()
                        if line != "" and not line.startswith("#"):
                            line = line.split("=")
                            if len(line) != 2:
                                raise CoverMiException("Malformed options file: {0}".format("=".join(line)))
                            if line[0] in self._options:
                                raise CoverMiException("Duplicate option: {0}".format(line[0]))
                            self._options[line[0]] = line[1]
            if "Depth" in self._options:
                if  not self._options["Depth"].isdigit():                
                    raise CoverMiException("Depth is not numeric: {0}".format(self._options["Depth"]))
                self._options["Depth"] = int(self._options["Depth"])
        return self._options


    def write_options(self):
        if self._options is None:
            self._options = {}
        if "Options" not in self.file_names:
            self.file_names["Options"] =  "covermi_options.txt"
            if os.path.lexists(os.path.join(self.panel_path, self.file_names["Options"])):
                raise CoverMiException("File covermi_options.txt exists but is not of correct format")
        with open(os.path.join(self.panel_path, self.file_names["Options"]), "wt") as f: 
            f.write("#CoverMi options\n")
            for key, value in self._options.items():
                f.write("{0}={1}\n".format(key, value))
        if "Depth" in self.file_names:
            os.unlink(os.path.join(self.panel_path, self.file_names["Depth"]))


    def __contains__(self, attr):
        try:
            test = getattr(self, attr)
        except CoverMiException:
            return False
        return True



if __name__ == "__main__":
    panel = Panel("/home/ed/Desktop/panels/Iron Panel v2 - 13.04.15_HD")
    print panel.variants_disease
    print panel.variants_mutation
    print panel.variants_gene
    print panel.depth



















import os


def load_conf():
    conf = {}
    try:
        with file(os.path.join(os.path.dirname(os.path.abspath(__file__)), "covermi.conf"), "rU") as f:
            for line in f:
                if "=" in line:
                    key, value = line.split("=")
                    conf[key.strip()] = value.strip()
    except IOError:
        pass
    if "panel_root" not in conf or not os.path.exists(conf["panel_root"]):
        conf["panel_root"] = os.path.dirname(os.path.expanduser("~"))
    if "bam_root" not in conf or not os.path.exists(conf["bam_root"]):
        conf["bam_root"] = os.path.dirname(os.path.expanduser("~"))
    return conf



import csv, pdb
from itertools import islice
from collections import defaultdict
from operator import attrgetter
import reader
try:
    from itertools import izip as zip
except ImportError:
    pass


class Gr(object):

    def __repr__(self):
        return repr(self.data)


    def __init__(self, data=None):
        self.data = defaultdict(list)
        self._tuple = None
        self._hash = None
        if data is not None:
            if not hasattr(data, "__iter__"):
                data = (data,)
            for entry in data:
                self.add(entry)
            self.sort()


    def add(self, entry):
        self.data[entry.chrom].append(entry)
        return self


    def sort(self):
        for chrom in self.data.values():
            chrom.sort(key=attrgetter("start", "stop"))


    @property
    def sorted_by_name(self):
        for entry in sorted([entry for entry in self], key=attrgetter("name")):
            yield entry


    @property
    def all_entries(self):
        return self.__iter__()


    def __iter__(self):
        for key ,chrom in sorted(self.data.items()):
            for entry in chrom:
                yield entry


    @property
    def as_tuple(self):
        if not self._tuple: self._tuple = tuple([(key, tuple([(entry.start, entry.stop) for entry in value])) for key, value in sorted(self.data.items())])
        return self._tuple


    @property
    def as_set(self):
        return set((entry.chrom, entry.start, entry.stop) for entry in self)


    def __hash__(self):
        if not self._hash: self._hash = hash(self.as_tuple)
        return self._hash


    def __eq__(self, other):
        return self.as_tuple == other.as_tuple


    @property
    def inverted(self):
        new = type(self)()
        for key in range(0, len(reader.CHR_2_STR)):
            last_stop = 0
            if self.data[key]:
                for entry in self.data[key]:
                    if last_stop+1 < entry.start-1:
                        new.add(entry._replace(start=last_stop+1, stop=entry.start-1))
                    last_stop = entry.stop
            else:
                entry = reader.Entry(key, None, None, "", reader.PLUS)
            new.add(entry._replace(start=last_stop+1, stop=2**29-1))
        return new


    @property
    def merged(self): # The name of a range will be the name of the first sub-range that was merged into it
        new = type(self)()
        for key, chrom in self.data.items():
            nchrom = [chrom[0]]
            for entry in islice(chrom, 1, len(chrom)):
                if entry.start-1 <= nchrom[-1].stop:
                    if entry.stop > nchrom[-1].stop:
                        nchrom[-1] = nchrom[-1]._replace(stop=entry.stop)
                else:
                    nchrom.append(entry)
            new.data[key] = nchrom
        return new


    def overlapped_by(self, other):# If range other contains overlapping ranges then we may get multiple copies of the overlapping ranges
        new = type(self)()
        for key, chrom in self.data.items():
            for a in chrom:
                if key in other.data:
                    for b in other.data[key]:
                        if b.start > a.stop:
                            break
                        if (a.start <= b.start <= a.stop) or (a.start <= b.stop <= a.stop) or b.stop > a.stop:
                            new.add(a._replace(start=max(a.start, b.start), stop=min(a.stop, b.stop)))
        new.sort() # Only needed if gr contains overlapping ranges which should not happen with sane use
        return new


    def not_touched_by(self, other):
        new = type(self)()
        for key, chrom in self.data.items():
            for a in chrom:
                touching = False
                if key in other.data:
                    for b in other.data[key]:
                        if b.start > a.stop:
                            break
                        if (a.start <= b.start <= a.stop) or (a.start <= b.stop <= a.stop) or b.stop > a.stop:
                            touching = True
                            break
                if not touching:
                    new.add(a)
        return new


    def touched_by(self, other):
        new = type(self)()
        for key, chrom in self.data.items():
            for a in chrom:
                if key in other.data:
                    for b in other.data[key]:
                        if b.start > a.stop:
                            break
                        if (a.start <= b.start <= a.stop) or (a.start <= b.stop <= a.stop) or b.stop > a.stop:
                            new.add(a)
                            break
        return new

    
    def subranges_covered_by(self, other):
        new = type(self)()
        for key, chrom in self.data.items():
            for a in chrom:
                if key in other.data:
                    for b in other.data[key]:
                        if b.start > a.start:
                            break
                        if b.stop >= a.stop:
                            new.add(a)
                            break
        return new


    def combined_with(self, other):
        new = type(self)()
        for gr in (self, other):
            for entry in gr:
                new.add(entry)
        new.sort()
        return new


    def subset(self, names, exclude=False, genenames=False):
        if not hasattr(names, "__iter__"):
            names = (names,)
        names = set([name.split()[0] for name in names] if genenames else names)
        new = type(self)()
        for entry in self:
            if ((entry.name.split()[0] if genenames else entry.name) in names) ^ exclude:
                new.add(entry)
        return new


    @property
    def names(self):
        return sorted(set([entry.name for entry in self]))


    @property
    def number_of_components(self):
        return sum([len(chrom) for chrom in self.data.values()])


    @property
    def number_of_weighted_components(self):
        return sum([entry.weight for entry in self])


    @property
    def base_count(self):
        return sum([entry.stop - entry.start + 1 for entry in self])


    @property
    def locations_as_string(self):
        return ", ".join(["{0}:{1}-{2}".format(reader.CHR_2_STR[entry.chrom], entry.start, entry.stop) for entry in self])


    @property
    def names_as_string(self):   ###########################???
        namedict = defaultdict(list)
        for entry in self:
            try:
                namedict[entry.name].append(entry.exon)
            except AttributeError:
                pass
        namelist = []
        for name, numbers in sorted(namedict.items()):
            numbers.sort()
            exons = []
            index = 0
            while index<=len(numbers)-1:
                start = numbers[index]
                stop = start
                for index2 in xrange(index+1, len(numbers)+1):
                    if index2 == len(numbers) or numbers[index2] != stop+1:
                        break
                    stop += 1
                exons.append("e{0}{1}".format(start, "" if (start==stop) else "-{0}".format(stop)))
                index = index2
            namelist.append("{0} {1}".format(name, ",".join(exons)).strip())
        return ", ".join(namelist)


    @property
    def is_empty(self):
        return self.number_of_components == 0


    def save(self, f): #Save type(self)() object in bedfile format, START POSITION IS ZERO BASED
        try:
            writer = csv.writer(f, delimiter="\t")
        except TypeError:
            with open(f, "wb") as realfile:
                self.save(realfile)
            return
        for entry in self:
            writer.writerow((reader.CHR_2_STR[entry.chrom], entry.start-1, entry.stop, entry.name, entry.strand))
            




from reportfunctions import TextTable, header, location
from gr import Gr
from reader import SPLICE_SITE_BUFFER
import pdb


def create(coverage, panel, identifier, outputstem):

    if "exons" not in panel or "transcripts" not in panel or "depth" not in panel or "amplicons" not in panel:
        return

    somatic = panel.options["ReportType"]=="Somatic" if ("ReportType" in panel.options) else False
    frequency = panel.options["VariantFrequency"]=="True" if ("VariantFrequency" in panel.options) else False
    report = header(panel, identifier)

    # Coverage by Gene
    table = TextTable()
    table.headers.append(["Gene", "Coverage"])
    for i in coverage.calculate(panel.exons,panel.depth):
        table.rows.append([i.name, (i.percent_covered, "{:.0f}%")])
    if len(table.rows) > 0:
        report += ["\n\n"] + table.formated(sep="    ")


    # Coverage by Gene (for genes not in panel)
    table = TextTable()
    table.headers.append(["Gene", "Coverage"])
    othergenes = panel.allexons.touched_by(panel.amplicons).subset(panel.transcripts.names, exclude=True, genenames=True)
    if not othergenes.is_empty:
        for i in coverage.calculate(othergenes,panel.depth):
            table.rows.append([i.name, (i.percent_covered, "{:.0f}%")])
    if len(table.rows) > 0:
        report += ["\n\n"] + ["Coverage of genes not in the targeted panel\n"]  + table.formated(sep="    ")
  

    # Coverage by Variant per Gene
    if "variants_gene" in panel:
        table = TextTable()
        table.headers.append(["Gene", "Variants Covered", "Clinical Sensitivity" if frequency else ""])
        for i in coverage.calculate(panel.variants_gene,panel.depth):
            table.rows.append([i.name, 
                              [i.components_covered, "/", i.components, "(", (i.percent_components_covered, "{:.0f}%)")],
                              (i.percent_weighted_components_covered, "{:.0f}%") if frequency else ""])
        if len(table.rows) > 0:
           report += ["\n\n"] + table.formated(sep="    ")


    # Coverage by Variant per Disease
    if "variants_disease" in panel:
        table = TextTable()
        table.headers.append(["Disease", "Variants Covered", "Clinical Sensitivity" if frequency else ""])
        for i in coverage.calculate(panel.variants_disease,panel.depth):
            table.rows.append([i.name,
                              [i.components_covered, "/", i.components, "(", (i.percent_components_covered,  "{:.0f}%)")],
                              (i.percent_weighted_components_covered, "{:.0f}%") if frequency else ""])
        if len(table.rows) > 0:
            report += ["\n\n"] + table.formated(sep="    ")


    # Coverage by Individual Variant
    if "variants_mutation" in panel:
        table = TextTable()
        table.headers.append(["Gene", "Mutation", "Location", "Proportion of" if frequency else "", "Disease"])
        table.headers.append(["", "", "", "Mutations in" if frequency else "", ""])
        table.headers.append(["", "", "", "Gene" if frequency else "", ""])
        weighted_mutations_per_gene = {}
        for entry in panel.variants_mutation.all_entries:
            gene = entry.name.split()[0]
            if gene not in weighted_mutations_per_gene:
                weighted_mutations_per_gene[gene] = 0
            weighted_mutations_per_gene[gene] += entry.weight
        for i in coverage.calculate(panel.variants_mutation,panel.depth):
            if (somatic and i.completely_covered) or (not somatic and i.incompletely_covered):
                table.rows.append([i.name.split()[0],
                                   i.name.split()[1],
                                   i.range_combined.locations_as_string,
                                   (float(i.weighted_components_covered if somatic else i.weighted_components_uncovered)*100/weighted_mutations_per_gene[i.name.split()[0]], "{:.2f}%") if frequency else "",
                                   i.diseases])
        if len(table.rows) > 0:
            report += ["\n\n"] + ["Variants "+("" if somatic else "not ")+"covered by panel\n"]
            report += table.formated(sep="  ", sortedby=3, reverse=True, trim_columns=((4, 20), (1,20))) if frequency else table.formated(sep="  ", trim_columns=((3, 20),))


    # Coverage by Exon
    table = TextTable()
    table.headers.append(["Exon", "Coverage", "Covered Region" if somatic else "Uncovered Region"])
    for i in coverage.calculate(panel.exons,panel.depth, exons=True):
        if (somatic and i.bases_covered>0) or (not somatic and i.bases_uncovered>0):
            table.rows.append([i.name, 
                               (i.percent_covered, "{:.0f}%"),
                               i.range_covered.locations_as_string if somatic else i.range_uncovered.locations_as_string])
    if len(table.rows) > 0:
        report += ["\n\n"] + ["Exons "+("" if somatic else "not fully ")+"covered by panel\n"] + table.formated(sep="  ")


    rogue_amplicons = panel.amplicons.not_touched_by(panel.transcripts)
    if not rogue_amplicons.is_empty:
        table = TextTable()
        table.headers.append(["Amplicon", "Location"])
        for entry in rogue_amplicons.all_entries:
            table.rows.append([entry.name, location(Gr().add(entry), panel)])
        if len(table.rows) > 0:
            report += ["\n\n"] + ["Amplicons not covering a targeted gene\n"] + table.formated(sep="    ")


#    if "ExcludedAmplicons" in panel:
#        table = TextTable()
#        table.headers.append(["Amplicon", "Location"])
#        for entry in panel("ExcludedAmplicons").all_entries:
#            table.rows.append([entry.name, location(Gr().add(entry), panel)])
#        if len(table.rows) > 0:
#            report += ["\n\n"] + ["Amplicons in manifest file that have been excluded from analysis\n"] + table.formated(sep="    ")


    table = TextTable()
    table.headers.append(["Exon", "Upstream Padding", "Downstream Padding"])
    for chr_name in panel.transcripts.data:
        for transcript in panel.transcripts.data[chr_name]:
            exons = panel.exons.touched_by(Gr().add(transcript)).subset(transcript.name)
            amplicons = panel.amplicons.touched_by(Gr().add(transcript)).merged
            loopover = range(0, len(exons.data[chr_name])) if (transcript.strand=="+") else range(len(exons.data[chr_name])-1, -1, -1)
            for index in loopover:
                touching_amplicons = amplicons.touched_by(Gr().add(exons.data[chr_name][index]))
                if touching_amplicons.is_empty:
                    continue
                prev_stop = exons.data[chr_name][index-1].stop if (index>0) else 0
                next_start = exons.data[chr_name][index+1].start if (index<len(exons.data[chr_name])-1) else 2**29#MAX_CHR_LENGTH
                prev_amp = touching_amplicons.data[chr_name][0].start
                next_amp = touching_amplicons.data[chr_name][-1].stop
                padding = [ exons.data[chr_name][index].start-max(prev_amp, prev_stop)+SPLICE_SITE_BUFFER,
                            min(next_amp, next_start)-exons.data[chr_name][index].stop+SPLICE_SITE_BUFFER ]
                
                table.rows.append([Gr().add(exons.data[chr_name][index]).names_as_string,
                                   padding[transcript.strand == "-"] if (padding[transcript.strand == "-"]>0) else "",
                                   padding[transcript.strand == "+"] if (padding[transcript.strand == "+"]>0) else ""])
    if len(table.rows) > 0:
        table.rows.sort()
        report += ["\n\n"] + table.formated(sep="  ")

    with file(outputstem+"_covermi_design_report.txt", "wt") as f:
        f.writelines(report)




import unittest
from gr import Gr
from reader import Entry, PLUS, MINUS




class TestGr(unittest.TestCase):

    def test_gr_add(self):
        one   = Entry(0, 10, 20, "", PLUS)
        two   = Entry(0, 21, 30, "", PLUS)
        three = Entry(1, 15, 25, "", PLUS)

        range1 = Gr().add(three)
        range1.add(two)
        range1.add(one)

        range1.sort()

        self.assertEqual(range1.number_of_components, 3)
        self.assertEqual([entry for entry in range1.all_entries], [one, two, three])


    def test_gr_reader(self):
        one   = Entry(0, 10, 20, "", PLUS)
        two   = Entry(0, 21, 30, "", PLUS)
        three = Entry(1, 15, 25, "", PLUS)

        range1 = Gr([three, two, one])

        self.assertEqual(range1.number_of_components, 3)
        self.assertEqual([entry for entry in range1.all_entries], [one, two, three])


    def test_gr_merge(self):
        one   = Entry(0, 10, 20, "", PLUS)
        two   = Entry(0, 21, 30, "", PLUS)
        three = Entry(0, 31, 40, "", PLUS)
        four  = Entry(0, 9, 41, "", PLUS)
        five  = Entry(0, 99, 100, "", PLUS)
   
        merged = Gr([three, two]).merged

        self.assertEqual(merged.number_of_components, 1)
        self.assertEqual([entry for entry in merged.all_entries], [Entry(0, 21, 40, "", PLUS)])

        merged = Gr([three, two, one]).merged

        self.assertEqual(merged.number_of_components, 1)
        self.assertEqual([entry for entry in merged.all_entries], [Entry(0, 10, 40, "", PLUS)])

        merged = Gr([four, three, two, one]).merged

        self.assertEqual(merged.number_of_components, 1)
        self.assertEqual([entry for entry in merged.all_entries], [Entry(0, 9, 41, "", PLUS)])

        merged = Gr([five, four, three, two, one]).merged

        self.assertEqual(merged.number_of_components, 2)
        self.assertEqual([entry for entry in merged.all_entries], [Entry(0, 9, 41, "", PLUS), five])







































if __name__ == '__main__':
    unittest.main()









#!/usr/bin env python
import sys, os, time, re, getopt, pdb
from cov import Cov
from panel import Panel
from files import Files
import clinicalreport, covermiplot
from covermiexception import CoverMiException


def create_output_dir(output_path):
    output_path = os.path.join(output_path, "covermi_output")
    try:
        os.mkdir(output_path)
    except OSError:
        if os.path.isdir(output_path):
            raise CoverMiException("{0} folder already exists".format(output_path))
        else:
            raise CoverMiException("Unable to create folder {0}".format(output_path))
    return output_path


def main(panel_path, bamlist, output_path):
    panel = Panel(panel_path).load()
    output_path = create_output_dir(output_path)
    print "Processing..."
    if len(set([os.path.splitext(os.path.basename(fn))[0].split("_")[0] for fn, depths in bamlist])) > len(bamlist):
        print "Duplicate sample numbers. Quiting"
        exit()

    starttime = time.time()
    for bam_path, depths in bamlist:
        panel["Options"]["Depths"] = depths
        panel["Filenames"]["Sample"] = os.path.splitext(os.path.basename(bam_path))[0].split("_")[0]
        cov = Cov.load_bam(bam_path, panel["Exons"], amplicons=False)
        for depth in depths:
            panel["Options"]["Depth"] = depth
            clinicalreport.create(cov, panel, os.path.join(output_path, "{0}_{1}x_".format(panel["Filenames"]["Sample"], depth)))
            seconds = int(time.time() - starttime)
            time_string = "{0} sec".format(seconds) if (seconds<60) else "{0} min {01} sec".format(seconds/60, seconds%60)
            print "File {}, Depth {}, {}".format(panel["Filenames"]["Sample"], depth, time_string)
        panel["Options"]["Depth"] = max(depths)
        covermiplot.plot(cov, panel, os.path.join(output_path, panel["Filenames"]["Sample"]))


#    try:
#        opts, args = getopt.getopt(sys.argv[1:], "p:b:o:", ["panel=", "bams=", "output="])
#    except getopt.GetoptError as err:
#        raise CoverMiException(str(err))

#    output = None
#    bams = None
#    panel = None
#    depth = None
#    for o, a in opts:
#        if o in ("-p", "--panel"):
#            panel = a
#        elif o in ("-b", "--bams"):
#            bams = a
#        elif o in ("-o", "--output"):
#            output = a
#        elif o in ("-d", "--depth"):
#            depth = a
#        else:
#            raise CoverMiException("Unrecognised option {0}".format(o))

#    main(panel, bams, output, depth)











#!/usr/bin env python
import sys, os, tkFileDialog, Tkinter, pdb, traceback, getopt
import covermimain, covermiconf
from panel import Panel


class DepthDialog(object):
    def __init__(self, parent, default_depth):
        self.parent = parent
        self.parent.depth = ""
        self.window = Tkinter.Toplevel(self.parent)
        self.window.title("CoverMi")
        Tkinter.Label(self.window, text="Please enter minimum depth").grid(column=0, row=0)
        self.entry = Tkinter.Entry(self.window)
        self.entry.grid(column=1, row=0)
        self.entry.insert(0, str(default_depth))
        self.entry.bind('<Return>', self.return_pressed)
        self.entry.focus_set()

    def return_pressed(self, event):
        if self.entry.get().isdigit():
            self.parent.depth = self.entry.get()
            self.window.destroy()


class M_S_D_Dialog(object):
    def __init__(self, parent):
        self.parent = parent
        self.parent.msd = ""
        self.window = Tkinter.Toplevel(self.parent)
        self.window.title("CoverMi")
        Tkinter.Label(self.window, text="Do you want to check a single bam, multiple bams or review the panel design?").grid(column=0, row=0, columnspan=3, padx=10, pady=5)
        Tkinter.Button(self.window, text="Folder of bam files", width=15, command=self.multiple_pressed).grid(column=0, row=1, pady=10)
        Tkinter.Button(self.window, text="Single bam file", width=15, command=self.single_pressed).grid(column=1, row=1, pady=10)
        Tkinter.Button(self.window, text="Review panel design", width=15, command=self.design_pressed).grid(column=2, row=1, pady=10)

    def single_pressed(self):
        self.parent.msd = "single"
        self.window.destroy()

    def multiple_pressed(self):
        self.parent.msd = "multiple"
        self.window.destroy()

    def design_pressed(self):
        self.parent.msd = "design"
        self.window.destroy()


class SomCon_Dialog(object):
    def __init__(self, parent):
        self.parent = parent
        self.parent.somcon = ""
        self.window = Tkinter.Toplevel(self.parent)
        self.window.title("CoverMi")
        Tkinter.Label(self.window, text="Is this a somatic or constitutional panel?").grid(column=0, row=0, columnspan=2, padx=10, pady=5)
        Tkinter.Button(self.window, text="Somatic", width=15, command=self.som_pressed).grid(column=0, row=1, pady=10, padx=10)
        Tkinter.Button(self.window, text="Constitutional", width=15, command=self.con_pressed).grid(column=1, row=1, pady=10, padx=10)

    def som_pressed(self):
        self.parent.somcon = "Somatic"
        self.window.destroy()

    def con_pressed(self):
        self.parent.somcon = "Constitutional"
        self.window.destroy()


def main():
    try:
        conf = covermiconf.load_conf()

        rootwindow = Tkinter.Tk()
        rootwindow.withdraw()

        print("Please select a panel")
        panelpath = tkFileDialog.askdirectory(parent=rootwindow, initialdir=conf["panel_root"], title='Please select a panel')
        if not bool(panelpath):
            sys.exit()
        panelpath = os.path.abspath(panelpath)
        print("{0} panel selected".format(os.path.basename(panelpath)))

        panel = Panel(panelpath)
        rootwindow.wait_window(DepthDialog(rootwindow, str(panel.depth) if ("depth" in panel) else "").window)
        if rootwindow.depth == "":
            sys.exit()
        panel.options["Depth"] = int(rootwindow.depth)
        print("Depth {0} selected".format(panel.depth))

        if "ReportType" not in panel.options:
            print("Is this a somatic or constitutional panel?")
            rootwindow.wait_window(SomCon_Dialog(rootwindow).window)    
            if rootwindow.somcon == "":
                sys.exit()
            panel.options["ReportType"] = rootwindow.somcon

        panel.write_options()

        print("Do you wish to coverage check multiple bams, a single bam or review the panel design?")
        rootwindow.wait_window(M_S_D_Dialog(rootwindow).window)    
        mode = str(rootwindow.msd)
        if mode == "":
            sys.exit()
        elif mode == "design":
            bampath = ""
            outputpath = conf["panel_root"]  
            print("Design review selected")
        else:
            if mode == "multiple":
                print("Please select the folder containing the bam files")
                bampath = tkFileDialog.askdirectory(parent=rootwindow, initialdir=conf["bam_root"], title='Please select a folder')
            elif mode == "single":
                print("Please select a bam file")
                bampath = tkFileDialog.askopenfilename(parent=rootwindow, initialdir=conf["bam_root"], filetypes=[("bamfile", "*.bam")], title='Please select a bam file')
            if bampath == "":
                sys.exit()
            bampath = os.path.abspath(bampath)
            outputpath = os.path.dirname(bampath) if (mode=="single") else bampath

            print("{0} selected".format(bampath))

        print("Please select a location for the output")
        outputpath = tkFileDialog.askdirectory(parent=rootwindow, initialdir=outputpath, title='Please select a location for the output')
        if outputpath == "":
            sys.exit()
        outputpath = os.path.abspath(outputpath)
        print("Output folder {0} selected".format(outputpath))

        covermimain.main(panelpath, bampath, outputpath)

        print("Finished")
    except Exception as e:
#        if type(e).__name__ == "CoverMiException":
#            print(e.message)
#        else:
            traceback.print_exc()
            print("UNEXPECTED ERROR. QUITTING.")
    finally:
        raw_input("Press any key to continue...")

if __name__ == "__main__":
    if getopt.getopt(sys.argv[1:], "", ["args"])[1]: # has arguments
        covermimain.main()
    else: # no arguments
        main()







import time, pdb
from version import __version__

class TextTable(object):

    def __init__(self):
        self.rows = []
        self.headers = []

    @classmethod
    def _aligned(cls, table, delete_empty_cols=True):
        columns = len(table[0])
        rows = len(table)
        newtab = [[] for n in range(0, rows)]

        for col in range(0, columns):
            if type(table[0][col]) == list:
                replacement = cls._aligned([table[row][col] for row in range(0, rows)])
                for row in range(0, rows):
                    newtab[row].append("".join(replacement[row]))
            else:
                if type(table[0][col]) == tuple:
                    fstring = table[0][col][1]
                    for row in range(0, rows):
                        table[row][col] = table[row][col][0]
                else:
                    fstring = "{}"

                biggest = max([len(fstring.format(table[row][col])) for row in range(0, rows)])
                if not (biggest == 0 and delete_empty_cols):
                    for row in range(0, rows):
                        newtab[row].append(fstring.format(table[row][col]).ljust(biggest) if isinstance(table[row][col], basestring) else fstring.format(table[row][col]).rjust(biggest))
        return newtab


    def formated(self, sep="", sortedby=None, reverse=False, page_width=99, trim_columns=()):
        if sortedby is not None:
            def bycolumn(line): return [line[sortedby]] + line
            self.rows.sort(key=bycolumn, reverse=reverse)

        rows = type(self)._aligned(self.rows, delete_empty_cols=(True if len(self.headers)==0 else False))
        if len(self.headers) > 0:
            rows = type(self)._aligned(self.headers + rows)

        for column_number, minimum_width in trim_columns:
            excess_width = len(sep.join(rows[0])) - page_width
            if excess_width > 0:
                current_column_width = len(rows[0][column_number])
                if current_column_width > minimum_width:
                    new_column_width = max(current_column_width - excess_width, minimum_width)
                    for row in rows:
                        if row[column_number][new_column_width:].strip() == "":
                            row[column_number] = row[column_number][0:new_column_width]
                        else:
                            row[column_number] = row[column_number][0:new_column_width-3]+"..."

        if len(self.headers) > 0:
            table = [sep.join(row)+"\n" for row in rows]
            table = [("-"*(len(table[0])-1))+"\n"] + table[0:len(self.headers)] + [("-"*(len(table[0])-1))+"\n"] + table[len(self.headers):]
        else:
            table = [sep.join(row)+"\n" for row in rows]
        return table


def header(panel, identifier):
    table = TextTable()
    if identifier is not None:
        table.rows.append(["Sample:", identifier.name])
        table.rows.append(["Run:", identifier.run])

    for descriptor, key in (("Panel: ", "Panel"), ("Manifest: ", "Manifest"), ("Design Studio Bedfile:", "DesignStudio"),
                             ("Known variants file: ", "Variants"), ("Panel type:", "ReportType"), ("Minimum Depth: ", "Depth")):
        for catagory in ("file_names", "options"):
            if key in getattr(panel, catagory):
                table.rows.append([descriptor, str(getattr(panel, catagory)[key])])
                break
    table.rows += [["Date of CoverMi analysis: ", time.strftime("%d/%m/%Y")], ["CoverMi version: ", __version__]]
    return table.formated()


def location(gr1, panel):
    if "exons" and "transcripts" in panel:
        exons = panel.allexons if ("allexons" in panel) else panel.exons
        transcripts = panel.alltranscripts if ("alltranscripts" in panel) else panel.transcripts
        loc = exons.overlapped_by(gr1).names_as_string
        if loc =="":
            loc = "{0} intron".format(transcripts.overlapped_by(gr1).names_as_string)
            if loc == " intron":
                loc = "not covering a "+("" if ("alltranscripts" in panel) else"targeted ")+"gene"
    else:
        loc = ""
    return loc




#!/usr/bin env python
import os, sys, tkFileDialog, Tkinter
from .gr import Gr, Entry
from .files import Files


def shared(proband=None, test=None, chromosome=None):
    if chromosome:
        proband = proband.touched_by(chromosome)
        test = test.touched_by(chromosome)
    p = set([entry.name for entry in proband.all_entries if entry.passfilter])
    t = set([entry.name for entry in test.all_entries if entry.passfilter])
    return "{:.1f}%".format(float(len(p & t)*100)/len(t)) if len(t)>0 else "NA"


def main():
    root_dir = os.path.dirname(os.path.abspath(__file__))
    rootwindow = Tkinter.Tk()
    rootwindow.withdraw()

    probandpath = tkFileDialog.askopenfilename(parent=rootwindow, initialdir=root_dir, title='Please select the proband vcf')
    if not bool(probandpath):
        sys.exit()
    print "proband: {}".format(os.path.basename(probandpath))

    testpath = tkFileDialog.askdirectory(parent=rootwindow, initialdir=root_dir, title='Please select the folder of the vcfs to be checked')
    if not bool(testpath):
        sys.exit()

    chrY = Gr().add(Entry("chrY", 1, Gr.MAX_CHR_LENGTH, "", "+"))
    chrM = Gr().add(Entry("chrM", 1, Gr.MAX_CHR_LENGTH, "", "+"))
    proband_variants = Gr.load_vcf(probandpath)

    for sample, run, testpath in Files(testpath, ".vcf"): 
        test_variants = Gr.load_vcf(testpath+".vcf")
        print "{} {}, common variants: {}, chrY: {}, chrM {}".format(
              os.path.basename(sample), 
              "male" if len(test_variants.touched_by(chrY))>0 else "female",
              shared(proband=proband_variants, test=test_variants), 
              shared(proband=proband_variants, test=test_variants, chromosome=chrY),
              shared(proband=proband_variants, test=test_variants, chromosome=chrM))

if __name__ == "__main__":
    main()



import sys, os, tkFileDialog, Tkinter, pdb


def main():
    rootwindow = Tkinter.Tk()
    rootwindow.withdraw()

    print("Please select knownCanonical file downloaded from UCSC")
    kcpath = tkFileDialog.askopenfilename(parent=rootwindow, filetypes=[("textfile", "*.txt")], title='Please select knownCanonical file downloaded from UCSC')
    if not bool(kcpath):
        sys.exit()
    print("{0} file selected".format(os.path.basename(kcpath)))
    path = os.path.dirname(kcpath)
    print("Please select kgXref file downloaded from UCSC")
    kgpath = tkFileDialog.askopenfilename(parent=rootwindow, filetypes=[("textfile", "*.txt")], initialdir=path, 
                                          title='Please select kgXref file downloaded from UCSC')
    if not bool(kgpath):
        sys.exit()
    print("{0} file selected".format(os.path.basename(kgpath)))


    canonicals = set([])
    with file(kcpath) as f:
        for line in f:
            chrom, start, stop, cluster, kgid, protein = line.split("\t")
            canonicals.add(kgid)

    path = os.path.join(path, "covermi_"+os.path.basename(kcpath))
    if os.path.exists(path):
        raise RuntimeError("File {} already exists".format(path))
    with file(path, "wt") as g:
        with file(kgpath) as f:
            for line in f:
                kgid, mrna, upan, upid, gene, transcript, protein, des, rfam, rna = line.split("\t")
                if kgid in canonicals and transcript != "":
                    g.write("{0}\t{1}\n".format(gene, transcript))

    print "Written file {}".format(path)
    print "This file needs to be placed in all panel directories"


if __name__ == "__main__":
    main()



#!/usr/bin env python
import os, tkFileDialog, Tkinter
from .gr import Gr, Entry


def notshared(proband=None, test=None, chromosome=None):
    if chromosome:
        proband = proband.touched_by(chromosome)
        test = test.touched_by(chromosome)
    p = set([entry.name for entry in proband.all_entries if entry.passfilter])
    t = set([entry.name for entry in test.all_entries if entry.passfilter])
    return "{:.1f}%".format(float(len(p - t)*100)/len(p)) if len(p)>0 else "NA"


def main():
    rootwindow = Tkinter.Tk()
    rootwindow.withdraw()

    probandpath = tkFileDialog.askopenfilename(parent=rootwindow, title='Please select the probands vcf')
    if not bool(probandpath):
        sys.exit()
    print "proband: {}".format(os.path.basename(probandpath))

    motherpath = tkFileDialog.askopenfilename(parent=rootwindow, title='Please select the mothers vcf')
    if not bool(motherpath):
        sys.exit()
    print "mother: {}".format(os.path.basename(motherpath))

    fatherpath = tkFileDialog.askopenfilename(parent=rootwindow, title='Please select the fathers vcf')
    if not bool(fatherpath):
        sys.exit()
    print "father: {}".format(os.path.basename(fatherpath))

    chrY = Gr().add(Entry("chrY", 1, Gr.MAX_CHR_LENGTH, "", "+"))
    chrM = Gr().add(Entry("chrM", 1, Gr.MAX_CHR_LENGTH, "", "+"))
    proband_variants = Gr.load_vcf(probandpath)
    mother_variants = Gr.load_vcf(motherpath)
    father_variants = Gr.load_vcf(fatherpath)

    print "{} of the probands variants are not present in the parents".format(
        notshared(proband=proband_variants, test=mother_variants.combined_with(father_variants)))
    print "{} of the probands variants on chromosome M are not present in the mother".format(
        notshared(proband_variants, mother_variants, chrM))
    print "{} of the probands variants on chromosome Y are not present in the father".format(
        notshared(proband_variants, father_variants, chrY))

if __name__ == "__main__":
    main()



# { chr_number : [ [start, stop, depth], [start, stop, depth] ] }

import pdb
from collections import defaultdict
from gr import Gr
import reader

PLUS = reader.PLUS
MINUS = reader.MINUS

MAX_CHR_LENGTH = 2**29 # Max length on a reference sequence in an indexed BAM + 1


def bisect_left(a, x):
    lo = 0
    hi = len(a)
    while lo < hi:
        mid = (lo+hi)//2
        if a[mid][1] < x.start: lo = mid+1
        else: hi = mid
    return lo


def fake_paired_end_reads(amplicons, depth):
    for entry in amplicons.all_entries:
        third = (entry.stop - entry.start + 1) // 3
        for n in range(0, depth):
            yield (entry.chrom, entry.start, entry.stop-third, MINUS if entry.strand==MINUS else PLUS)
            yield (entry.chrom, entry.start+third, entry.stop, PLUS if entry.strand==MINUS else MINUS)


def extend(data, item):
    if len(data) and item[2] == data[-1][2]:
        data[-1][1] = item[1]
    else:
        data += [item]


def normalised(data, depth):
    newdata = defaultdict(zero_coverage)
    for key, values in data.items():
        newchrom = []
        for entry in values:
            extend(newchrom, [entry[0], entry[1], entry[2]>=depth])
        newdata[key] = newchrom
    return newdata


def combined(mydata, otherdata):
    newdata = defaultdict(zero_coverage)
    for key in set(mydata.keys()+otherdata.keys()):
         newchrom = []
         otherchrom = otherdata[key].__iter__()
         otherentry = [None, -1, None]
         for myentry in mydata[key]:
             while myentry[0] > otherentry[1]:
                 otherentry = otherchrom.next()
             while otherentry[1] < myentry[1]:
                 extend(newchrom, [max(myentry[0], otherentry[0]), otherentry[1], myentry[2]+otherentry[2]])
                 otherentry = otherchrom.next()
             extend(newchrom,[max(myentry[0], otherentry[0]), myentry[1], myentry[2]+otherentry[2]])
         newdata[key] = newchrom
    return newdata


class CumCov(object):

    def __init__(self, depth):
        self._data = defaultdict(zero_coverage)
        self.number = 0
        self.depth = depth


    def add(self, other):
        self.number += 1
        self._data = combined(self._data, normalised(other.data, self.depth))
        return self


    @property
    def data(self):
        newdata = defaultdict(zero_coverage)
        for key, chrom in self._data.items():
            newdata[key] = [(entry[0], entry[1], entry[2]*100/self.number) for entry in chrom]
        return newdata


def zero_coverage(): return [[0, MAX_CHR_LENGTH, 0]]


class Cov(object):

    def __init__(self, data, amplicons=(), locations=(), print_progress=False):
        self.data = defaultdict(zero_coverage)

        if isinstance(data, basestring):
            bam = reader.Bam(data, index=bool(locations))
            data = bam.coverage(locations=locations, print_progress=print_progress)
        else:
            bam = None

        fr_depth = {}
        self.amplicon_info = []
        self.ontarget = 0
        self.offtarget = 0
        for amplicon in amplicons:
            amplicon_info = AmpliconInfo(amplicon)
            self.amplicon_info += [amplicon_info]
            fr_depth[(amplicon.start, PLUS)] = amplicon_info
            fr_depth[(amplicon.stop, MINUS)] = amplicon_info

        for chrom, start, stop, strand in data:
            depth = 1
            cov = self.data[chrom]
            upperx = -1
            lowerx = -1
            for x in xrange(len(cov) - 1, -1, -1):
                if stop < cov[x][0]:
                    continue
                if stop >= cov[x][1] and start <= cov[x][0]:
                    cov[x][2] += depth
                    if start == cov[x][0]:
                        break
                else:
                    if stop < cov[x][1]:
                        upperx = x
                    if start > cov[x][0]:
                        lowerx = x
                        break
            if upperx != -1:
                cov.insert(upperx, [cov[upperx][0], stop, cov[upperx][2] + depth])
                cov[upperx+1][0] = stop + 1
            if lowerx != -1:
                cov.insert(lowerx+1, [start, cov[lowerx][1], cov[lowerx][2] + depth * (upperx!=lowerx)])
                cov[lowerx][1] = start - 1
                if upperx == lowerx:
                    cov[lowerx][2] -= depth

            if amplicons:
                try:
                    fr_depth[(start, PLUS)].f_depth += 1
                    self.ontarget += 1
                except KeyError:
                    try:
                        fr_depth[(stop, MINUS)].r_depth += 1
                        self.ontarget += 1
                    except KeyError:
                        self.offtarget += 1                        

        self.amplicon_info.sort()
        if bam:
            self.unmapped = bam.unmapped
            self.mapped = bam.mapped


    def load(self, path):
        self.__init__()
        with open(path, "rb") as f:
            for line in csv.reader(f, delimiter="\t"):
                self.data[line[0]].append([int(value) for value in line[1:4]])


    def save(self, path):
        with open(path, "wb") as f:
            writer = csv.writer(f, delimiter="\t")
            for chrom in self:
                for entry in self[chrom]:
                    writer.writerow([chrom]+entry)


    def calculate(self, gr1, min_depth, exons=False, total=False):
        results = defaultdict(CoverageInfo)
        for entry in gr1.all_entries:
            name = entry.name if (not total) else ""
            name = name if (not exons) else "{0} e{1}".format(name, entry.exon)
            info = results[name]
            info.name = name
            info.diseases = getattr(entry, "diseases", "")

            allcovered = True
            cchrom = self.data[entry.chrom]
            bisect = bisect_left(cchrom, entry) # leftmost coverage.stop >= entry.start
            for cstart, cstop, cdepth in cchrom[bisect:]:
                if cstart > entry.stop:
                    break
                elif cstop >= entry.start:
                    bases = min(entry.stop, cstop) - max(entry.start, cstart) + 1
                    if cdepth>=min_depth:
                        info.bases_covered += bases
                        info.depth_covered += bases * cdepth
                        info.range_covered.add(entry._replace(start=max(entry.start, cstart), stop=min(entry.stop, cstop)))
                    else:
                        info.bases_uncovered += bases
                        info.depth_uncovered += bases * cdepth
                        info.range_uncovered.add(entry._replace(start=max(entry.start, cstart), stop=min(entry.stop, cstop)))
                        allcovered = False
            if allcovered:
                info.components_covered += 1
                info.weighted_components_covered += getattr(entry, "weight", 0)
            else:
                info.components_uncovered += 1
                info.weighted_components_uncovered += getattr(entry, "weight", 0)

        results = sorted(results.values())
        for info in results:
            info.depth_covered /= max(info.bases_covered, 1)
            info.depth_uncovered /= max(info.bases_uncovered, 1)
            info.range_covered = info.range_covered.merged
            info.range_uncovered = info.range_uncovered.merged
        return results if (not total) else results[0]


class AmpliconInfo(object):

    def __repr__(self):
        return "{} forward={} reverse={}".format(self.entry, self.f_depth, self.r_depth)

    def __init__(self, entry):
        self.entry = entry
        self.f_depth = 0
        self.r_depth = 0

    def __lt__(self, other):
        return self.entry < other.entry

    @property
    def name(self):
        return self.entry.name

    @property
    def chrom(self):
        return reader.CHR_2_STR[self.entry.chrom]

    @property
    def minimum_depth(self):
        return min(self.f_depth, self.r_depth)

    @property
    def maximum_depth(self):
        return max(self.f_depth, self.r_depth)

    @property
    def mean_depth(self):
        return (self.f_depth + self.r_depth)/2

    @property
    def ratio(self):
        return float(self.minimum_depth)/max(self.maximum_depth, 1)

    @property
    def gr(self):
        return Gr().add(self.entry)


class CoverageInfo(object):

    def __repr__(self):
        return "{} {}%".format(self.name, self.percent_covered)

    def __init__(self):
        self.name = ""
        self.diseases = ""
        self.depth_covered = 0
        self.depth_uncovered = 0
        self.bases_covered = 0
        self.bases_uncovered = 0
        self.range_covered = Gr()
        self.range_uncovered = Gr()
        self.components_covered = 0
        self.components_uncovered = 0
        self.weighted_components_covered = 0
        self.weighted_components_uncovered = 0

    def __lt__(self, other):
        return self.name < other.name
    
    @property
    def depth(self):
        return ((self.depth_covered * self.bases_covered) + (self.depth_uncovered + self.bases_uncovered)) / self.bases

    @property
    def percent_covered(self):
        return float(self.bases_covered*100) / self.bases

    @property
    def percent_uncovered(self):
        return 100 - self.percent_covered

    @property
    def range_combined(self):
        return self.range_covered.combined_with(self.range_uncovered).merged

    @property
    def bases(self):
        return self.bases_covered + self.bases_uncovered

    @property
    def components(self):
        return self.components_covered + self.components_uncovered

    @property
    def percent_components_covered(self):
        return float(self.components_covered*100) / self.components

    @property
    def percent_components_uncovered(self):
        return 100 - self.percent_components_covered

    @property
    def weighted_components(self):
        return self.weighted_components_covered + self.weighted_components_uncovered

    @property
    def percent_weighted_components_covered(self):
        return float(self.weighted_components_covered*100) / self.weighted_components

    @property
    def percent_weighted_components_uncovered(self):
        return 100 - self.percent_weighted_components_covered

    @property
    def completely_covered(self):
        return not(self.incompletely_covered)

    @property
    def incompletely_covered(self):
        return bool(self.bases_uncovered)



from reportfunctions import TextTable, header
from gr import Gr
import pdb
import os
import time


def textable(body):
    return "\\begin{tabular}{ " + " ".join(["l"]*len(body[0]))+"}\n" + \
           "\\\\\n".join([" & ".join([str(item)for item in row]) for row in body]) + \
           "\\\\\n\\end{tabular}\n"

def texbigtable(headings, body):
    return 





def create(coverage, panel, outputstem):

    tex = [r"""
\documentclass[12pt]{article}
\usepackage{underscore}

\title{CoverMi Report}

\begin{document}
\maketitle
"""]


    if "Exons" not in panel or "Transcripts" not in panel or "Depth" not in panel["Options"]:
        return

    frequency = panel["Options"]["VariantFrequency"]=="True" if ("VariantFrequency" in panel["Options"]) else False

    minimum_depth = panel["Options"]["Depth"]
    if "Amplicons" in panel:
        targeted_range = panel["Amplicons"].merged
        targeted_exons = panel["Exons"].overlapped_by(targeted_range)
    else:
        targeted_range = panel["Transcripts"]
        targeted_exons = panel["Exons"]


    body = []
    for descriptor, item in (("Sample: ", "Sample"), ("Run: ", "Run"), ("Panel: ", "Panel"), ("Manifest: ", "Manifest"), ("Design Studio Bedfile:", "DesignStudio"),
                             ("Known variants file: ", "Variants"), ("Panel type:", "ReportType"), ("Minimum Depth: ", "Depth")):
        for catagory in ("Filenames", "Options"):
            if item in panel[catagory]:
                body += [[descriptor, str(panel[catagory][item])]]
                break
    body += [["CoverMi version: ", "XXX"]]
    tex += [textable(body)]
        

    # Total Coverage
    i = coverage.calculate(targeted_exons, minimum_depth, total=True)
    tex += ["\\subsection*{Overall Coverage}\n"] + \
           ["{0} of {1} bases ({2:0.1f}\\%) covered with a mean depth of {3}\n".format(i.bases_covered, i.bases, i.percent_covered, i.depth_covered)]


 #    # Coverage by Gene
    tex += ["\\subsection*{Coverage by Gene}\n"]
    headers = ["Gene", "Coverage of Targeted Region", "Coverage of Whole Gene", "Mean Depth"]
    body = []
    for i in coverage.calculate(targeted_exons, minimum_depth):
        body += [[i.name, "{}".format(i.percent_covered), "{}".format(i.bases_covered*100.0/panel["Exons"].subset(i.name).base_count), i.depth_covered]]
    if body:
        tex += [textable(body)]




















#    # Summary Table for WGS
#    if "Depths" in panel["Options"]:
#        table = TextTable()
#        table.headers.append(["Depth", "Proportion of genes with"])
#        table.headers.append(["",      "at least 90% coverage"])

#        for depth in sorted(panel["Options"]["Depths"], reverse=True):
#            info = coverage.calculate(targeted_exons, depth)
#            table.rows.append([depth, (float(sum([int(i.percent_covered>=90) for i in info]))*100/len(info), "{:.0f}%")])
#        if len(table.rows) > 0:
#            report += ["\n\n"] + table.formated(sep="    ")

#    # Coverage by Variant per Gene
#    if "Variants_Gene" in panel:
#        table = TextTable()
#        table.headers.append(["Gene", "Variants Covered", "Variants Covered", "Clinical" if frequency else ""])
#        table.headers.append(["", "in Targeted Region", "in Whole Gene     ", "Sensitivity" if frequency else ""])
#        targeted_variants = panel["Variants_Gene"].subranges_covered_by(targeted_range)
#        for i in coverage.calculate(panel["Variants_Gene"], minimum_depth):
#            detectable = targeted_variants.subset(i.name).number_of_components
#            table.rows.append([i.name, 
#                              [i.components_covered, "/", detectable, "(", (float(i.components_covered)*100/max(detectable,1), "{:.0f}%)")],
#                              [i.components_covered, "/", i.components, "(", (i.percent_components_covered, "{:.0f}%)")],
#                              (i.percent_weighted_components_covered, "{:.0f}%") if frequency else ""])
#        if len(table.rows) > 0:
#           report += ["\n\n"] + table.formated(sep="    ")
#        

#    # Coverage by Variant per Disease
#    if "Variants_Disease" in panel:
#        table = TextTable()
#        table.headers.append(["Disease", "Variants Covered", "Variants Covered", "Clinical" if frequency else ""])
#        table.headers.append(["", "in Targeted Region", "in Whole Geneome  ", "Sensitivity" if frequency else ""])
#        targeted_variants = panel["Variants_Disease"].subranges_covered_by(targeted_range)
#        for i in coverage.calculate(panel["Variants_Disease"], minimum_depth):
#            detectable = targeted_variants.subset(i.name).number_of_components
#            table.rows.append([i.name,
#                              [i.components_covered, "/", detectable , "(", (float(i.components_covered*100)/max(detectable,1), "{:.0f}%)")],
#                              [i.components_covered, "/", i.components, "(", (i.percent_components_covered,  "{:.0f}%)")],
#                              (i.percent_weighted_components_covered, "{:.0f}%") if frequency else ""])
#        if len(table.rows) > 0:
#            report += ["\n\n"] + table.formated(sep="    ")


#    # Coverage by Individual Variant
#    if "Variants_Mutation" in panel:
#        table = TextTable()
#        table.headers.append(["Gene", "Mutation", "Location", "Depth", "Proportion of" if frequency else "", "Disease"])
#        if frequency:
#            table.headers.append(["", "", "", "", "Mutations in" if frequency else "", ""])
#            table.headers.append(["", "", "", "", "Gene" if frequency else "", ""])
#        weighted_mutations_per_gene = {}
#        for entry in panel["Variants_Mutation"].all_entries:
#            gene = entry.name.split()[0]
#            if gene not in weighted_mutations_per_gene:
#                weighted_mutations_per_gene[gene] = 0
#            weighted_mutations_per_gene[gene] += entry.weight
#        for i in coverage.calculate(panel["Variants_Mutation"], minimum_depth):
#            if i.incompletely_covered:
#                table.rows.append([i.name.split()[0],
#                                   i.name.split()[1],
#                                   i.range_combined.locations_as_string,
#                                   i.depth_uncovered,
#                                   (float(i.weighted_components_uncovered)*100/weighted_mutations_per_gene[i.name.split()[0]], "{:.2f}%") if frequency else "",
#                                   i.diseases])
#        if len(table.rows) > 0:
#            report += ["\n\n"] + ["Inadequately covered targeted variants\n"] 
#            report += table.formated(sep="  ", sortedby=4, reverse=True, trim_columns=((5, 20), (1, 20))) if frequency else table.formated(sep="  ", trim_columns=((4, 20),))


#    # Coverage by Exon
#    table = TextTable()
#    table.headers.append(["Exon", "Coverage of", "Location of", "Mean Depth of"])
#    table.headers.append(["", "Targeted Region", "Uncovered Region", "Uncovered Region"])
#    for i in coverage.calculate(targeted_exons, minimum_depth, exons=True):
#        if i.bases_uncovered > 0:
#            table.rows.append([i.name, 
#                               (i.percent_covered, "{:.0f}%"),
#                               i.range_uncovered.locations_as_string, 
#                               i.depth_uncovered])
#    if len(table.rows) > 0:
#        report += ["\n\n"] + ["Inadequately covered targeted exons\n"] + table.formated(sep="  ")



    tex += [r"""\end{document}
"""]



    
    with file(outputstem+".tex", "wt") as f:
        f.writelines(tex)


import reader
from panel import Panel
from cov import Cov
import pickle

if __name__ == "__main__":

    panel = Panel("/home/ed/Desktop/bampy/Iron Panel v2 - 13.04.15_HD").load()
    #cov = Cov(data=reader.Bam("/home/ed/Desktop/bampy/G165527L_S3.bam").coverage())
    #with open("cov.pickle", "wb") as f:
    #    pickle.dump(cov, f)
    with open("cov.pickle", "rb") as f:
        cov = pickle.load(f)
    create(cov, panel, os.path.join("./textest"))








#!/usr/bin env python
from panel import Panel
import sys, os, tkFileDialog, Tkinter, re, csv
import pdb



def translate(row):
    if row["Primary site"] == "haematopoietic_and_lymphoid_tissue":
        try:
            return row["Histology subtype 1"]
        except KeyError:
            return row["Histology subtype"] 
    return ""



def main():
    chr_translate = {str(x): "chr{}".format(y) for x, y in zip(range(1, 26), range(1, 23)+["X", "Y", "M"])}

    rootwindow = Tkinter.Tk()
    rootwindow.withdraw()

    print "Please select the COSMIC mutation tsv file"
    cosmic_path = tkFileDialog.askopenfilename(parent=rootwindow, title='Please select the COSMIC mutation file')
    if cosmic_path == "":
        sys.exit()
    with file(cosmic_path, "rU") as f:
        if  "Gene name" not in f.readline().split("\t"):
            abort("File {0} is of incorrect format".format(os.path.basename(cosmic_path)))

    print "Please select a panel"
    panel_path = tkFileDialog.askdirectory(parent=rootwindow, title='Please select a panel')
    if panel_path == "":
        sys.exit()

    panel = Panel(panel_path)
    if "Variants" in panel:
        abort("Variants file {0} already exists within panel".format(os.path.basename(panel["Variants"])))
    if "Targets" not in panel:
        abort("Unable to identify target gene file")
    if "Disease_Names" in panel:
        print "Disease names file {0} will be deleted and updated".format(os.path.basename(panel["Disease_Names"]))
        if raw_input("Press y to continue, any other key to abort") != "y":
            sys.exit()
    target_genes = set(panel.load()["Transcripts"].names)

    outputfilepath = os.path.join(panel_path, 
                              "{0}_{1}.txt".format(os.path.basename(panel_path.rstrip(os.pathsep)), os.path.splitext(os.path.basename(cosmic_path.rstrip(os.pathsep)))[0]))
    if os.path.exists(outputfilepath):
        abort("File {0} already exists".format(outputfilepath))

    print "Writing file {0}".format(os.path.basename(outputfilepath))
    with file(outputfilepath, "wb") as of:
        writer = csv.writer(of, delimiter="\t")
        writer.writerow(["HGMD ID", "Disease", "Variant Class", "Gene Symbol", "chromosome", "coordinate start", "coordinate end", "strand", "hgvs"])

        diseases = {}
        with file(cosmic_path ,"rU") as f:
            for row in csv.DictReader(f, delimiter="\t"):
                gene = row["Gene name"].split("_")[0]
                if gene in target_genes:
                    disease = translate(row)
                    if disease and row["Mutation genome position"]:
                        diseases[disease] = ""
                        chrom, startstop = row["Mutation genome position"].split(":")
                        start, stop = startstop.split("-")
                        writer.writerow([ ".", disease, ".", gene, chr_translate[chrom], start, stop, row["Mutation strand"], row["Mutation AA"]])

        with file(cosmic_path ,"rU") as f:
            for row in csv.DictReader(f, delimiter="\t"):
                gene = row["Gene name"].split("_")[0]
                if gene not in target_genes:
                    disease = translate(row)
                    if disease and row["Mutation genome position"]:
                        if disease in diseases:
                            chrom, startstop = row["Mutation genome position"].split(":")
                            start, stop = startstop.split("-")
                            writer.writerow([ ".", disease, ".", gene, chr_translate[chrom], start, stop, row["Mutation strand"], row["Mutation AA"]])

    if "Disease_Names" in panel:
        with file(panel["Disease_Names"]) as f:
            for l in f:
                if l[0] != "#":
                    old_name, new_name = l.strip().split("=")
                    if old_name in diseases:
                        diseases[old_name] = "="+new_name
        os.unlink(panel["Disease_Names"])

    diseasefilepath = "{0}_Disease_Names.txt".format(os.path.splitext(outputfilepath)[0])
    if os.path.exists(diseasefilepath):
        abort("File {0} already exists".format(os.path.basename(diseasefilepath)))

    print "Writing file {0}".format(os.path.basename(diseasefilepath))
    with file(diseasefilepath, "wt") as f:
        f.write("#Variants Disease Name Translation\n")
        for t in sorted(diseases.items()):
            f.write("".join(t)+"\n")    

    options = panel.read_options()
    options["VariantFrequency"] = "True"
    panel.write_options(options)



def abort(message):
    print message
    raw_input("Press any key to quit")
    sys.exit()



if __name__ == "__main__":
    main()






try:
    from matplotlib.backends.backend_pdf import FigureCanvasPdf, PdfPages
    from matplotlib.figure import Figure
    from matplotlib.patches import Rectangle
    from matplotlib.backends.backend_pdf import PdfPages
    MATPLOTLIB = True
except ImportError:
    MATPLOTLIB = False


from math import log10
from itertools import izip, cycle

from gr import Gr
from reader import PLUS, MINUS

import pdb



class Plot(object):
    def __init__(self, cumulative=False):
        self.cumulative = cumulative
        self.data = []
        self.new()

    def new(self, *args):
        self.x = []
        self.y = []
        if args: self.add(*args)

    def add(self, x, y):
        self.x += [self.scalex(x)]
        self.y += [self.scaley(y)]

    def scalex(self, x):
        for beginning, end, fixed, scaling in self.data:
            if x >= beginning and x <= end:
                break
        rel_pos = x - beginning
        return x - (fixed + (rel_pos - (rel_pos/scaling)))

    def scaley(self, y):
        return (log10(y) if y > 1 else 0.0) if not self.cumulative else y / 25.0


def plot(coverage, panel, identifier, fn, cumulative=False):
    if "transcripts" not in panel or not MATPLOTLIB:
        return

    with PdfPages(fn) as pdf:
        for entry in panel.transcripts.sorted_by_name:
            figure = Figure(figsize=(11.7, 4.15))
            FigureCanvasPdf(figure)
            ax = figure.gca()

            transcript = Gr().add(entry)
            amplicons = panel.amplicons.touched_by(transcript) if ("amplicons" in panel) else Gr()
            exons = panel.exons.touched_by(transcript).subset(entry.name) if ("exons" in panel) else Gr()
            plot_area = transcript.combined_with(amplicons).merged
            plot_start = plot_area.data[entry.chrom][0].start
            plot_stop = plot_area.data[entry.chrom][0].stop

            max_intron_size = amplicons.base_count / amplicons.number_of_components if not amplicons.is_empty else 200
            blocks = amplicons.combined_with(exons).merged
            # data = [start, stop, fixed subtraction, scaling factor]
            extra = {"cumulative": True} if cumulative else {}
            plot = Plot(**extra)
            end_of_prev_block = plot_start-2
            fixed_total = 0
            for block in blocks:
                intron_size = block.start - end_of_prev_block - 1
                plot.data += [(end_of_prev_block+1, block.start-1, fixed_total, max(float(intron_size)/max_intron_size, 1))]
                fixed_total += max(intron_size - max_intron_size, 0)
                plot.data += [(block.start, block.stop, fixed_total, 1)]
                end_of_prev_block = block.stop
            plot.data += [(end_of_prev_block+1, plot_stop+1, fixed_total, max(float(plot_stop - end_of_prev_block)/max_intron_size, 1) if (not blocks.is_empty) else 1)]

            # coverage
            plot.new(plot_start-1, 0)
            for cstart, cstop, cdepth in coverage.data[entry.chrom]:
                if cstop >= plot_start:
                    plot.add(max(plot_start, cstart), cdepth)
                    if cstop >= plot_stop:
                        plot.add(plot_stop, cdepth)
                        break
            plot.add(plot_stop+1, 0)
            ax.plot(plot.x, plot.y, "-", drawstyle="steps-post", color="dodgerblue", linewidth=1)

            # amplicons
            for amplicon, y in izip(amplicons, cycle((-0.04, 0))):
                ax.add_patch(Rectangle((plot.scalex(amplicon.start), y), amplicon.stop-amplicon.start+1, 0.04, edgecolor="black", facecolor="black", zorder=100))

            # exons
            xlabels = []
            xlocations = []
            for exon in exons:
                ax.add_patch(Rectangle((plot.scalex(exon.start), -0.3), exon.stop-exon.start+1, 0.2, edgecolor="black", facecolor="dimgrey"))
                xlabels += [exon.exon]
                xlocations += [plot.scalex((exon.start+exon.stop)/2)]
            ax.axhline(-0.2, color="black", linewidth=2, zorder=0)
            step = (len(xlocations) // 20) + 1
            ax.set_xticks(xlocations[::step])
            ax.set_xticklabels(xlabels[::step], fontsize=8)
            ax.tick_params(axis="x",length=0)

            # variants       
            if "variants_mutation" in panel:
                if cumulative:
                    yvar = 50
                elif "depth" in panel:
                    yvar = panel.depth
                else:
                    yvar = 0
                x = [plot.scalex((variant.start+variant.stop)/2) for variant in plot_area.overlapped_by(panel.variants_mutation)]
                y = [plot.scaley(yvar)] * len(x)
                ax.plot(x, y, "x", color="black")

            # depth        
            if "depth" in panel and not cumulative:
                ax.axhline(plot.scaley(panel.depth), color="black", linewidth=0.5, linestyle=":")

            border = (plot.scalex(plot_stop) - plot_start + 3) * 6/100
            minx = plot_start - 1 - border
            maxx = plot.scalex(plot_stop) + 1 + border
            if entry.strand == MINUS:
                minx, maxx = (maxx, minx)

            ax.set_xlim(minx, maxx)
            ax.spines["bottom"].set_visible(False)

            ax.set_ylim(-0.3, 4.5)
            ax.set_yticks([0, 1, 2, 3, 4])
            if cumulative:
                ax.set_yticklabels(["0%", "25%", "50%", "75%", "100%"], fontsize=8)
                ax.set_ylabel("Proportion of time covered")       
            else:
                ax.set_yticklabels(["0", "10", "100", "1000", "10,000"], fontsize=8)
                ax.set_ylabel("Read Depth (log scale)", fontsize=10)              
            ax.set_xlabel("Exon", fontsize=10)
            ax.set_title(identifier, fontsize=12)
            ax.add_patch(Rectangle((minx, 4.25), maxx-minx, 0.25, edgecolor="black", facecolor="bisque", zorder=100))
            ax.text((minx+maxx)/2, 4.355, entry.name, zorder=101, ha="center", va="center")
            
            pdf.savefig(figure)#, bbox_inches='tight')
            





#!/usr/bin env python
#!/usr/bin env python
import sys, os, tkFileDialog, Tkinter, pdb, traceback
import covermiconf
from panel import Panel


class DepthDialog(object):
    def __init__(self, parent, default_depth):
        self.parent = parent
        self.parent.depth = ""
        self.window = Tkinter.Toplevel(self.parent)
        self.window.title("CoverMi")
        Tkinter.Label(self.window, text="Please enter minimum depth").grid(column=0, row=0)
        self.entry = Tkinter.Entry(self.window)
        self.entry.grid(column=1, row=0)
        self.entry.insert(0, str(default_depth))
        self.entry.bind('<Return>', self.return_pressed)
        self.entry.focus_set()

    def return_pressed(self, event):
        if self.entry.get().isdigit():
            self.parent.depth = self.entry.get()
            self.window.destroy()


class M_S_D_Dialog(object):
    def __init__(self, parent):
        self.parent = parent
        self.parent.msd = ""
        self.window = Tkinter.Toplevel(self.parent)
        self.window.title("CoverMi")
        Tkinter.Label(self.window, text="Do you want to check a single bam, multiple bams or review the panel design?").grid(column=0, row=0, columnspan=3, padx=10, pady=5)
        Tkinter.Button(self.window, text="Folder of bam files", width=15, command=self.multiple_pressed).grid(column=0, row=1, pady=10)
        Tkinter.Button(self.window, text="Single bam file", width=15, command=self.single_pressed).grid(column=1, row=1, pady=10)
        Tkinter.Button(self.window, text="Review panel design", width=15, command=self.design_pressed).grid(column=2, row=1, pady=10)

    def single_pressed(self):
        self.parent.msd = "single"
        self.window.destroy()

    def multiple_pressed(self):
        self.parent.msd = "multiple"
        self.window.destroy()

    def design_pressed(self):
        self.parent.msd = "design"
        self.window.destroy()


class SomCon_Dialog(object):
    def __init__(self, parent):
        self.parent = parent
        self.parent.somcon = ""
        self.window = Tkinter.Toplevel(self.parent)
        self.window.title("CoverMi")
        Tkinter.Label(self.window, text="Is this a somatic or constitutional panel?").grid(column=0, row=0, columnspan=2, padx=10, pady=5)
        Tkinter.Button(self.window, text="Somatic", width=15, command=self.som_pressed).grid(column=0, row=1, pady=10, padx=10)
        Tkinter.Button(self.window, text="Constitutional", width=15, command=self.con_pressed).grid(column=1, row=1, pady=10, padx=10)

    def som_pressed(self):
        self.parent.somcon = "Somatic"
        self.window.destroy()

    def con_pressed(self):
        self.parent.somcon = "Constitutional"
        self.window.destroy()


def guimain():
    try:
        conf = covermiconf.load_conf()

        rootwindow = Tkinter.Tk()
        rootwindow.withdraw()

        print("Please select a panel")
        panelpath = tkFileDialog.askdirectory(parent=rootwindow, initialdir=conf["panel_root"], title='Please select a panel')
        if not bool(panelpath):
            sys.exit()
        panelpath = os.path.abspath(panelpath)
        print("{0} panel selected".format(os.path.basename(panelpath)))

        panel = Panel(panelpath)
        options = panel.read_options()
        rootwindow.wait_window(DepthDialog(rootwindow, str(options["Depth"]) if ("Depth" in options) else "").window)
        if rootwindow.depth == "":
            sys.exit()
        options["Depth"] = int(rootwindow.depth)
        print("Depth {0} selected".format(options["Depth"]))

        if "ReportType" not in options:
            print("Is this a somatic or constitutional panel?")
            rootwindow.wait_window(SomCon_Dialog(rootwindow).window)    
            if rootwindow.somcon == "":
                sys.exit()
            options["ReportType"] = rootwindow.somcon

        panel.write_options(options)

        print("Do you wish to coverage check multiple bams, a single bam or review the panel design?")
        rootwindow.wait_window(M_S_D_Dialog(rootwindow).window)    
        mode = str(rootwindow.msd)
        if mode == "":
            sys.exit()
        elif mode == "design":
            bampath = ""
            outputpath = conf["panel_root"]  
            print("Design review selected")
        else:
            if mode == "multiple":
                print("Please select the folder containing the bam files")
                bampath = tkFileDialog.askdirectory(parent=rootwindow, initialdir=conf["bam_root"], title='Please select a folder')
            elif mode == "single":
                print("Please select a bam file")
                bampath = tkFileDialog.askopenfilename(parent=rootwindow, initialdir=conf["bam_root"], filetypes=[("bamfile", "*.bam")], title='Please select a bam file')
            if bampath == "":
                sys.exit()
            bampath = os.path.abspath(bampath)
            outputpath = os.path.dirname(bampath) if (mode=="single") else bampath

            print("{0} selected".format(bampath))

        print("Please select a location for the output")
        outputpath = tkFileDialog.askdirectory(parent=rootwindow, initialdir=outputpath, title='Please select a location for the output')
        if outputpath == "":
            sys.exit()
        outputpath = os.path.abspath(outputpath)
        print("Output folder {0} selected".format(outputpath))

        main(panelpath, bampath, outputpath, "old")
        main(panelpath, bampath, outputpath, "python")

        print("Finished")
    except Exception as e:
        if type(e).__name__ == "CoverMiException":
            print(e.message)
        else:
            traceback.print_exc()
            print("UNEXPECTED ERROR. QUITTING.")
    finally:
        raw_input("Press any key to continue...")


import sys, os, time, re, getopt, pdb
from cov import Cov
from panel import Panel
from files import Files
import covermiconf, technicalreport, clinicalreport, designreport, covermiplot
#import testreport as clinicalreport
from covermiexception import CoverMiException


def create_output_dir(output_path, bam_path, suffix):
    output_path = os.path.join(output_path, "{0}_covermi_output_{1}".format(os.path.splitext(os.path.basename(bam_path))[0], suffix))
    try:
        os.mkdir(output_path)
    except OSError:
        if os.path.isdir(output_path):
            raise CoverMiException("{0} folder already exists".format(output_path))
        else:
            raise CoverMiException("Unable to create folder {0}".format(output_path))
    return output_path


def main(panel_path, bam_path, output_path, test, depth=None):

    if not os.path.exists(panel_path):
        panel_path = os.path.join(covermiconf.load_conf["panel_root"], panel_path)

    panel = Panel(panel_path).load(bam_path=="")
    if depth is not None:
        panel["Options"]["Depth"] = int(depth)
    output_path = create_output_dir(output_path, bam_path if bam_path!="" else panel_path, test)
    print "Processing..."

    if bam_path != "":
        bam_file_list = Files(bam_path, ".bam")
        if len(bam_file_list) == 1:
            clinical_report_path = output_path
            technical_report_path = output_path
        else:
            clinical_report_path = os.path.join(output_path, "clinical")
            technical_report_path = os.path.join(output_path, "technical")
            os.mkdir(clinical_report_path)
            os.mkdir(technical_report_path)

        output_stems = set([])
        for panel["Filenames"]["Run"], panel["Filenames"]["Sample"], path in bam_file_list:
            path += ".bam"
            start_time = time.time()
            print "{0}/{1}".format(panel["Filenames"]["Run"], panel["Filenames"]["Sample"])

            output_stem = panel["Filenames"]["Sample"]
            dup_num = 1
            while output_stem in output_stems:
                output_stem = "{0}({1})".format(panel["Filenames"]["Sample"], dup_num)
                dup_num += 1
            output_stems.add(output_stem)

            if test == "python":
                cov = Cov.load_bam_python(path, panel["Amplicons"])            
            elif test == "old":
                if "Amplicons" in panel:
                    cov = Cov.load_bam(path, panel["Amplicons"], amplicons=True)
                    technicalreport.create(cov.amplicon_info, panel, os.path.join(technical_report_path, output_stem))
                else:
                    cov = Cov.load_bam(path, panel["Exons"], amplicons=False)
            else:
                raise CoverMiException("Hard coding error!")
            
            clinicalreport.create(cov, panel, os.path.join(clinical_report_path, output_stem))
            covermiplot.plot(cov, panel, os.path.join(clinical_report_path, output_stem))
            seconds = int(time.time() - start_time)
            time_string = "{0} sec".format(seconds) if (seconds<60) else "{0} min {01} sec".format(seconds/60, seconds%60)
            print"file {0} of {1} completed in {2}".format(len(output_stems), len(bam_file_list), time_string)
    else:
        cov = Cov.perfect_coverage(panel["Amplicons"])
        designreport.create(cov, panel, os.path.join(output_path, panel["Filenames"]["Panel"]))
        covermiplot.plot(cov, panel, os.path.join(output_path, panel["Filenames"]["Panel"]))


if __name__ == "__main__":

    try:
        opts, args = getopt.getopt(sys.argv[1:], "p:b:o:", ["panel=", "bams=", "output="])
    except getopt.GetoptError as err:
        raise CoverMiException(str(err))

    output = None
    bams = None
    panel = None
    depth = None
    for o, a in opts:
        if o in ("-p", "--panel"):
            panel = a
        elif o in ("-b", "--bams"):
            bams = a
        elif o in ("-o", "--output"):
            output = a
        elif o in ("-d", "--depth"):
            depth = a
        else:
            raise CoverMiException("Unrecognised option {0}".format(o))

    main(panel, bams, output, "python", depth)
    main(panel, bams, output, "old", depth)










from reportfunctions import TextTable, header
from gr import Gr
import pdb


def create(coverage, panel, identifier, outputstem):

    if "exons" not in panel or "transcripts" not in panel or "depth" not in panel:
        return

    frequency = panel.options["VariantFrequency"]=="True" if ("VariantFrequency" in panel.options) else False

    if "amplicons" in panel:
        targeted_range = panel.amplicons.merged
        targeted_exons = panel.exons.overlapped_by(targeted_range)
    else:
        targeted_range = panel.transcripts
        targeted_exons = panel.exons

    report = header(panel, identifier)
        

    # Total Coverage
    i = coverage.calculate(targeted_exons, panel.depth, total=True)
    report += ["\n\n", "{0} of {1} bases ({2:0.1f}%) covered with a mean depth of {3}\n".format(i.bases_covered, i.bases, i.percent_covered, i.depth_covered)]


#    # Summary Table for WGS
#    if "Depths" in panel["Options"]:
#        table = TextTable()
#        table.headers.append(["Depth", "Proportion of genes with"])
#        table.headers.append(["",      "at least 90% coverage"])

#        for depth in sorted(panel["Options"]["Depths"], reverse=True):
#            info = coverage.calculate(targeted_exons, depth)
#            table.rows.append([depth, (float(sum([int(i.percent_covered>=90) for i in info]))*100/len(info), "{:.0f}%")])
#        if len(table.rows) > 0:
#            report += ["\n\n"] + table.formated(sep="    ")


    # Coverage by Gene
    table = TextTable()
    table.headers.append(["Gene", "Coverage of", "Coverage of", "Mean Depth"])
    table.headers.append(["", "Targeted Region", "Whole Gene     ", ""])
    for i in coverage.calculate(targeted_exons, panel.depth):
        table.rows.append([i.name, (i.percent_covered, "{:.0f}%"), (float(i.bases_covered*100)/panel.exons.subset(i.name).base_count, "{:.0f}%"), i.depth_covered])
    if len(table.rows) > 0:
        report += ["\n\n"] + table.formated(sep="    ")


    # Coverage by Variant per Gene
    if "variants_gene" in panel:
        table = TextTable()
        table.headers.append(["Gene", "Variants Covered", "Variants Covered", "Clinical" if frequency else ""])
        table.headers.append(["", "in Targeted Region", "in Whole Gene     ", "Sensitivity" if frequency else ""])
        targeted_variants = panel.variants_gene.subranges_covered_by(targeted_range)
        for i in coverage.calculate(panel.variants_gene, panel.depth):
            detectable = targeted_variants.subset(i.name).number_of_components
            table.rows.append([i.name, 
                              [i.components_covered, "/", detectable, "(", (float(i.components_covered)*100/max(detectable,1), "{:.0f}%)")],
                              [i.components_covered, "/", i.components, "(", (i.percent_components_covered, "{:.0f}%)")],
                              (i.percent_weighted_components_covered, "{:.0f}%") if frequency else ""])
        if len(table.rows) > 0:
           report += ["\n\n"] + table.formated(sep="    ")
        

    # Coverage by Variant per Disease
    if "variants_disease" in panel:
        table = TextTable()
        table.headers.append(["Disease", "Variants Covered", "Variants Covered", "Clinical" if frequency else ""])
        table.headers.append(["", "in Targeted Region", "in Whole Geneome  ", "Sensitivity" if frequency else ""])
        targeted_variants = panel.variants_disease.subranges_covered_by(targeted_range)
        for i in coverage.calculate(panel.variants_disease, panel.depth):
            detectable = targeted_variants.subset(i.name).number_of_components
            table.rows.append([i.name,
                              [i.components_covered, "/", detectable , "(", (float(i.components_covered*100)/max(detectable,1), "{:.0f}%)")],
                              [i.components_covered, "/", i.components, "(", (i.percent_components_covered,  "{:.0f}%)")],
                              (i.percent_weighted_components_covered, "{:.0f}%") if frequency else ""])
        if len(table.rows) > 0:
            report += ["\n\n"] + table.formated(sep="    ")


    # Coverage by Individual Variant
    if "variants_mutation" in panel:
        table = TextTable()
        table.headers.append(["Gene", "Mutation", "Location", "Depth", "Proportion of" if frequency else "", "Disease"])
        if frequency:
            table.headers.append(["", "", "", "", "Mutations in" if frequency else "", ""])
            table.headers.append(["", "", "", "", "Gene" if frequency else "", ""])
        weighted_mutations_per_gene = {}
        for entry in panel.variants_mutation.all_entries:
            gene = entry.name.split()[0]
            if gene not in weighted_mutations_per_gene:
                weighted_mutations_per_gene[gene] = 0
            weighted_mutations_per_gene[gene] += entry.weight
        for i in coverage.calculate(panel.variants_mutation, panel.depth):
            if i.incompletely_covered:
                table.rows.append([i.name.split()[0],
                                   i.name.split()[1],
                                   i.range_combined.locations_as_string,
                                   i.depth_uncovered,
                                   (float(i.weighted_components_uncovered)*100/weighted_mutations_per_gene[i.name.split()[0]], "{:.2f}%") if frequency else "",
                                   i.diseases])
        if len(table.rows) > 0:
            report += ["\n\n"] + ["Inadequately covered targeted variants\n"] 
            report += table.formated(sep="  ", sortedby=4, reverse=True, trim_columns=((5, 20), (1, 20))) if frequency else table.formated(sep="  ", trim_columns=((4, 20),))


    # Coverage by Exon
    table = TextTable()
    table.headers.append(["Exon", "Coverage of", "Location of", "Mean Depth of"])
    table.headers.append(["", "Targeted Region", "Uncovered Region", "Uncovered Region"])
    for i in coverage.calculate(targeted_exons, panel.depth, exons=True):
        if i.bases_uncovered > 0:
            table.rows.append([i.name, 
                               (i.percent_covered, "{:.0f}%"),
                               i.range_uncovered.locations_as_string, 
                               i.depth_uncovered])
    if len(table.rows) > 0:
        report += ["\n\n"] + ["Inadequately covered targeted exons\n"] + table.formated(sep="  ")

    
    with file(outputstem+"_covermi_clinical_report.txt", "wt") as f:
        f.writelines(report)




#!/usr/bin env python
import sys, os, tkFileDialog, Tkinter, pdb, traceback, pkg_resources

from cov import Cov
from panel import Panel
from files import Files
from itertools import izip
from covermimain import CoverMiException
import covermiconf, technicalreport, clinicalreport


def main():
    try:
        version = pkg_resources.require("CoverMi")[0].version

        rootwindow = Tkinter.Tk()
        rootwindow.withdraw()

        print "CoverMi version {}".format(version)
        print("Please select the verification directory. This must contain a single panel and one or more bams run with that panel.")
        path = tkFileDialog.askdirectory(parent=rootwindow, title='Please select the verification directory')
        if not bool(path):
            sys.exit()
        path = os.path.abspath(path)
        print("{0} directory selected".format(os.path.basename(path)))

        for root, dirnames, filenames in os.walk(path):
            break
        
        panel = None
        for dirname in dirnames:
            try:
                panel = Panel(os.path.join(root, dirname)).load()
                break
            except CoverMiException:
                pass
        if panel is None:
            raise CoverMiException("No panel within verification directory.")
        if "Amplicons" not in panel:
            raise CoverMiException("No amplicon information within panel.")

        bams = Files(path, ".bam")
        suffixes = ("_amp_covermi_clinical_report.txt", "_noamp_covermi_clinical_report.txt", "_covermi_technical_report.txt", "_amplicon_data.tsv")

        matched = 0
        for run, sample, path in bams:
            matched += sum([os.path.exists(os.path.join(root, sample+suffix)) for suffix in suffixes])

        if matched == 0:
            verify = ""
            print "Creating baseline verification files for CoverMi version {}".format(version)
        elif matched == len(suffixes) * len(bams):
            verify = "_"
            print "Verifying CoverMi version {}".format(version)
        else:
            raise CoverMiException("Incomplete number of verification files.")

        for run, sample, path in bams:
            panel["Filenames"]["Run"] = "verification"
            panel["Filenames"]["Sample"] = sample
            print "Processing test bam {}.".format(sample)
            path += ".bam"
            cov = Cov.load_bam(path, panel["Amplicons"], amplicons=True)
            technicalreport.create(cov.amplicon_info, panel, os.path.join(root, verify+sample))
            clinicalreport.create(cov, panel, os.path.join(root, verify+sample+"_amp"))

            cov = Cov.load_bam(path, panel["Exons"], amplicons=False)
            clinicalreport.create(cov, panel, os.path.join(root, verify+sample+"_noamp"))

            if verify:
                for fn in [sample+suffix for suffix in suffixes]:
                    with file(os.path.join(root, fn), "rt") as old:
                        with file(os.path.join(root, verify+fn), "rt") as new:
                            for new_row, old_row in izip(new, old):
                                if new_row != old_row:
                                    if not (new_row.startswith("Date of CoverMi analysis:") or new_row.startswith("CoverMi version:")):
                                        raise CoverMiException("VERIFICATIO FAILED in file {}\nThe following lines do not match.\n{}{}".format(fn, new_row, old_row))
                    print "{} verified.".format(fn)
                    os.unlink(os.path.join(root, verify+fn))



    except Exception as e:
        if type(e).__name__ == "CoverMiException":
            print(e.message)
        else:
            traceback.print_exc()
            print("UNEXPECTED ERROR. QUITTING.")
    finally:
        raw_input("Press any key to continue...")

if __name__ == "__main__":
    main()



#!/usr/bin env python
import sys, os, tkFileDialog, Tkinter, pdb, traceback
import covermiwgsmain
from panel import Panel


class DepthDialog(object):
    def __init__(self, parent, default_depth):
        self.parent = parent
        self.parent.depth = ""
        self.window = Tkinter.Toplevel(self.parent)
        self.window.title("CoverMi")
        Tkinter.Label(self.window, text="Please enter a comma separated list of depths").grid(column=0, row=0)
        self.entry = Tkinter.Entry(self.window)
        self.entry.grid(column=1, row=0)
        self.entry.insert(0, str(default_depth))
        self.entry.bind('<Return>', self.return_pressed)
        self.entry.focus_set()

    def return_pressed(self, event):
        depths = [depth.strip() for depth in self.entry.get().split(",")]
        if all([depth.isdigit() for depth in depths]):
            self.parent.depths = [int(depth) for depth in depths]
            self.window.destroy()


class YesNo_Dialog(object):
    def __init__(self, parent):
        self.parent = parent
        self.parent.yesno = ""
        self.window = Tkinter.Toplevel(self.parent)
        self.window.title("CoverMi")
        Tkinter.Label(self.window, text="Do you wish to coverage check any more bams?").grid(column=0, row=0, columnspan=2, padx=10, pady=5)
        Tkinter.Button(self.window, text="Yes", width=15, command=self.yes_pressed).grid(column=0, row=1, pady=10, padx=10)
        Tkinter.Button(self.window, text="No", width=15, command=self.no_pressed).grid(column=1, row=1, pady=10, padx=10)

    def yes_pressed(self):
        self.parent.yesno = "yes"
        self.window.destroy()

    def no_pressed(self):
        self.parent.yesno = "no"
        self.window.destroy()


def main():

    try:        
        root_dir = os.path.dirname(os.path.expanduser("~"))
        rootwindow = Tkinter.Tk()
        rootwindow.withdraw()

        print("Please select a panel")
        panelpath = tkFileDialog.askdirectory(parent=rootwindow, initialdir=root_dir, title='Please select a panel')
        if not bool(panelpath):
            sys.exit()
        print("{0} panel selected".format(os.path.basename(panelpath)))

        bamlist = []
        while True:
                print("Please select a bam file")
                bampath = tkFileDialog.askopenfilename(parent=rootwindow, initialdir=root_dir, filetypes=[("bamfile", "*.bam")], title='Please select a bam file')
                if bampath == "":
                    sys.exit()
                root_dir = os.path.dirname(bampath)
                print("{0} selected".format(bampath))

                rootwindow.wait_window(DepthDialog(rootwindow, "").window)
                if rootwindow.depths == "":
                    sys.exit()
                depths = rootwindow.depths
                print("Depth {0} selected".format(", ".join([str(depth) for depth in depths])))

                bamlist += [(bampath, depths)]

                rootwindow.wait_window(YesNo_Dialog(rootwindow).window)    
                if rootwindow.yesno == "":
                    sys.exit()
                elif rootwindow.yesno == "no":
                    break

        print("Please select a location for the output")   
        outputpath = tkFileDialog.askdirectory(parent=rootwindow, initialdir=root_dir, title='Please select a location for the output')
        if outputpath == "":
            sys.exit()
        print("Output location {0} selected".format(outputpath))

        covermiwgsmain.main(panelpath, bamlist, outputpath)

        print("Finished")
    except Exception as e:
        if type(e).__name__ == "CoverMiException":
            print(e.message)
        else:
            traceback.print_exc()
            print("UNEXPECTED ERROR. QUITTING.")
    finally:
        raw_input("Press any key to continue...")

if __name__ == "__main__":
    main()



from gr import Gr
import reader
from cov import Cov, CumCov
from panel import Panel, Panels
import clinicalreport, designreport, technicalreport, reportfunctions
from reader import Vcf, Bam, Bed, IlluminaManifest, Variants, RefFlat

from .version import __version__
from covermiexception import CoverMiException



from setuptools import setup

setup(name="covermi",
    version= "2.1.0",
    description="Coverage checking for next generation sequencing panels",
    url="http://github.com/eawilson/covermi",
    author="Ed Wilson",
    author_email="edwardadrianwilson@yahoo.co.uk",
    license="MIT",
    packages=["covermi"],
    include_package_data=True,
    zip_safe=True,
    entry_points={"console_scripts" : ["covermi=covermi.covermigui:main",
                                       "covermicompare=covermi.covermicompare:guimain",
                                       "covermiverifygui=covermi.covermiverifygui:main", 
                                       "covermiwgs=covermi.covermiwgsgui:main",
                                       "covermi_make_canonical=covermi.make_canonical:main",
                                       "bringmiup=covermi.bringmiup:main",
                                       "cosmic2variants=covermi.cosmic2variants:main",
                                       "familycheck=covermi.familycheck:main",
                                       "whosthedaddy=covermi.whosthedaddy:main"]}
)




from __future__ import print_function
import csv, copy, struct, os, sys, pdb
from itertools import izip, izip_longest, count, ifilter, chain, repeat
from collections import defaultdict, namedtuple, Counter
from io import BufferedReader
from covermiexception import CoverMiException

try:
    import Bio.bgzf
    BGZF = True
except ImportError:
    import gzip
    BGZF = False

SPLICE_SITE_BUFFER = 5

CHR_2_STR = ["chr1",
             "chr2",
             "chr3",
             "chr4",
             "chr5",
             "chr6",
             "chr7",
             "chr8",
             "chr9",
             "chr10",
             "chr11",
             "chr12",
             "chr13",
             "chr14",
             "chr15",
             "chr16",
             "chr17",
             "chr18",
             "chr19",
             "chr20",
             "chr21",
             "chr22",
             "chrX",
             "chrY",
             "chrM",
            ]

STR_2_CHR = {"chr1": 0, "1": 0,
             "chr2": 1, "2": 1,
             "chr3": 2, "3": 2,
             "chr4": 3, "4": 3,
             "chr5": 4, "5": 4,
             "chr6": 5, "6": 5,
             "chr7": 6, "7": 6,
             "chr8": 7, "8": 7,
             "chr9": 8, "9": 8,
             "chr10": 9, "10": 9,
             "chr11": 10, "11": 10,
             "chr12": 11, "12": 11,
             "chr13": 12, "13": 12,
             "chr14": 13, "14": 13,
             "chr15": 14, "15": 14,
             "chr16": 15, "16": 15,
             "chr17": 16, "17": 16,
             "chr18": 17, "18": 17,
             "chr19": 18, "19": 18,
             "chr20": 19, "20": 19,
             "chr21": 20, "21": 20,
             "chr22": 21, "22": 21,
             "chrX": 22, "chr23": 22, "23": 22, "X": 22,
             "chrY": 23, "chr24": 23, "24": 23, "Y": 23,
             "chrM": 24, "chr25": 24, "25": 24, "M": 24,
            }

PLUS = "+"
MINUS = "-"

class FileContext(object):
    def __init__(self, fn_or_obj, mode=None):
        self.fn_or_obj = fn_or_obj
        self.mode = mode

    def __enter__(self):
        if isinstance(self.fn_or_obj, basestring):
            self.fp = open(self.fn_or_obj, self.mode)
            return self.fp
        else:
            self.fn_or_obj.seek(0)
            return self.fn_or_obj

    def __exit__(self, type_, value, traceback):
        if hasattr(self, "fp"):
            self.fp.close()


Entry = namedtuple("Entry", "chrom start stop name strand")
#VcfEntry = namedtuple("VcfEntry", "chrom start stop name strand indel quality vaf passfilter")
TranscriptEntry = namedtuple("TranscriptEntry", "chrom start stop name strand transcript")
ExonEntry = namedtuple("ExonEntry", "chrom start stop name strand exon")
VariantEntry = namedtuple("VariantEntry", "chrom start stop name strand weight")
MutationEntry = namedtuple("MutationEntry", "chrom start stop name strand weight diseases")

class VcfEntryClass(object):
    def __repr__(self):
        return "(chrom={}, start={}, stop={}, name={}, strand={}, indel={}, quality={}, vaf={}, passfilter={})".format(
                 self.chrom, self.start, self.stop, self.name, self.strand, self.indel, self.quality, self.vaf, self.passfilter)

    def __init__(self, chrom, start, stop, name, strand, indel, quality, vaf, passfilter):
        self.chrom = chrom
        self.start = start
        self.stop = stop
        self.name = name
        self.strand = strand
        self.indel = indel
        self.quality = quality
        self.vaf = vaf
        self.passfilter = passfilter

    def __eq__(self, other):
        return self.chrom == other.chrom and self.start == other.start and self.stop == other.stop and self.name == other.name

    def __hash__(self):
        return hash((self.chrom, self.start, self.stop, self.name))

CHROM = 0
POS = 1
SNPID = 2
REF = 3
ALTS = 4
QUAL = 5
FILTERS = 6
JUNK = 7
HEADINGS = 8
VALUES = 9

class Vcf(object):
    def __init__(self, path):
        self.path = path


    def __iter__(self):
        return self.read()


    def read(self):
        with FileContext(self.path, "rb") as f:
            for line in csv.reader(f, delimiter="\t"):
                if line[0].startswith("#"):
                    continue
                
                chrom = STR_2_CHR[line[CHROM]]
                quality = line[QUAL]
                quality = None if quality=="." else float(quality)
                passfilter = line[FILTERS]
                passfilter = None if passfilter == "." else all([code in ("PASS", "LowVariantFreq") for code in passfilter.split(";")])

                try:
                    ad = [float(depth) for depth in line[VALUES].split(":")[line[HEADINGS].split(":").index("AD")].split(",")]
                    dp = max(sum(ad), 1)
                except (ValueError, IndexError):
                    ad = ()
                #print(line[REF], line[ALTS], "ad "+str(ad[1:]))
                for start, ref, alt, depth in izip(repeat(int(line[POS])), repeat(line[REF]), line[ALTS].split(","), ad[1:]):

#                    while ref[:2] == alt[:2]:
#                        ref = ref[1:]
#                        alt = alt[1:]
#                        start += 1

#                    while ref[-1] == alt[-1] and len(ref) > 1 and len(alt) > 1:
#                        ref = ref[:-1]
#                        alt = alt[:-1]

                    length =  len(alt) - len(ref)
                    if length == 0: # snp
                        stop = start
                    elif length > 0: # insertion
                        stop = start + 1
                    else: # deletion
                        stop = start + 1 + length
                    name = "{0}:{1} {2}>{3}".format(CHR_2_STR[chrom], start, ref, alt)
#                    if length != 0:
#                        print("{} indel={} pass={}".format(name, length!=0, passfilter))

                    yield VcfEntryClass(chrom, start, stop, name, PLUS,
                                   length!=0,  quality, depth / dp if depth else None, passfilter)

    @property
    def variant_caller(self):
        with FileContext(self.path, "rb") as f:
            for line in f:
                if line.startswith("##source="):
                    return line.strip()[9:]
                if not line.startswith("#"):
                    return ""


# Bedfile format 
#    0.  Chromosome
#    1.  Start (zero based)
#    2.  End   (one based)
#    3.  Name
#    4.  Score
#    5.  Strand


CHROM = 0
START = 1
STOP = 2
NAME = 3
STRAND = 5

class Bed(object):
    def __init__(self, path):
        self.path = path


    def __iter__(self):
        return self.read()


    def read(self):
        with FileContext(self.path, "rb") as f:
            for line in csv.reader(f, delimiter="\t"):
                yield Entry(STR_2_CHR[line[CHROM]], int(line[START])+1, int(line[STOP]), line[NAME] if len(line)>NAME else ".",
                            MINUS if (len(line)>STRAND and line[STRAND]==MINUS) else PLUS)


# Ilumina Manifest file format:
#    [Header]
#
#    [Probes]
#    0.  Name-ID
#    1.  Region-ID
#    2.  Target-ID 
#    3.  Species
#    4.  Build-ID
#    5.  Chromosome
#    6.  Start position (including primer, zero based)
#    7.  End position (including primer, zero based)
#    8.  Strand
#    9.  Upstream primer sequence (len() will give primer length)
#    10  Upstream hits
#    11. Downstream  primer sequence (len() will give primer length)
#
#    [Targets]
#    0.  Target-ID
#    1.  Target-ID
#    2.  Target number (1 = on target, >1 = off target)
#    3.  Chromosome
#    4.  Start position (including primer, zero based)
#    5.  End position (including primer, zero based)
#    6.  Probe strand
#    7.  Sequence (in direction of probe strand)


HEADER = 1
PROBES = 2
TARGETS = 3

class IlluminaManifest(object):
    def __init__(self, path, trim_probes=True):
        self.path = path
        self.trim_probes = trim_probes


    def __iter__(self):
        return self.read()


    def read(self):
        self._offtarget = []
        sections = defaultdict(int, (("Header", HEADER), ("Probes", PROBES), ("Targets", TARGETS)))
        section = HEADER
        heading_row = False
        probes = {}
        rename_offtarget = {}
        with FileContext(self.path, "rb") as f:
            for line in csv.reader(f, delimiter="\t"):

                if heading_row:
                    col = {heading: index for index, heading in enumerate(line)}
                    heading_row = False

                elif line[0].startswith("["):
                    section = sections[line[0][1:line[0].index("]")]]
                    heading_row = section!=HEADER

                elif section == PROBES:
                    probes[line[col["Target ID"]]] = (len(line[col["ULSO Sequence"]]), len(line[col["DLSO Sequence"]])) if self.trim_probes else (0, 0)

                elif section == TARGETS:
                    chrom = STR_2_CHR[line[col["Chromosome"]]]
                    name = line[col["TargetA"]]
                    start = int(line[col["Start Position"]])
                    stop = int(line[col["End Position"]])
                    strand = line[col["Probe Strand"]]
                    target_number = line[col["Target Number"]]

                    if line[col["Target Number"]] == "1":
                        standardised_name = "{0}:{1}-{2}".format(CHR_2_STR[chrom], start, stop)
                        rename_offtarget[name] = standardised_name
                    else: # Off-target
                        standardised_name = rename_offtarget[name]

                    entry = Entry(chrom, start+probes[name][strand=="+"], stop-probes[name][strand=="-"], standardised_name, MINUS if strand=="-" else PLUS)
                    if target_number == "1": # On-target
                        yield entry
                    else:
                        self._offtarget += [entry]

    def offtarget(self):
        for entry in self._offtarget:
            yield entry


def refflatreader(f, needed, canonicaldict, doublecheck):
    for line in csv.reader(f, delimiter="\t"):
        if line[2] not in STR_2_CHR:
            continue
        if needed:
            if line[0] in needed:
                needed[line[0]] = False
            elif line[1] in needed and line[0] == doublecheck[line[1]]:
                needed[line[1]] = False
            else:
                continue
        else:
            if line[0] in canonicaldict and line[1] != canonicaldict[line[0]]:
                continue
        yield line


class RefFlat(object):
    TRANSCRIPTS = 0
    EXONS = 1

    def __init__(self, path, canonical=(), genes=(), splice_site_buffer=SPLICE_SITE_BUFFER):
        self.path = path
        self.splice_site_buffer = splice_site_buffer
        canonicaldict = {} # {gene: transcript}
        for line in csv.reader(canonical, delimiter="\t"):
            canonicaldict[line[0]] = line[1].split(" ")[-1]

        needed = {} # [gene or transcript]
        doublecheck = {} # {transcript: gene}
        for line in genes:
            line = " ".join(line.split()).split() # [gene, transcript if it exists]
            if len(line):
                if len(line) == 1: # gene OR transcript only 
                    if line[0] in canonicaldict:
                        needed[canonicaldict[line[0]]] = True
                        doublecheck[canonicaldict[line[0]]] = line[0]
                    else:
                        needed[line[0]] = True
                else:
                    needed[line[1]] = True
                    doublecheck[line[1]] = line[0]
        self.canonicaldict = canonicaldict
        self.doublecheck = doublecheck
        self.needed = needed
    
        with FileContext(path, "rb") as f:
            copies = Counter()
            self.n_transcripts = Counter()
            for line in refflatreader(f, needed, self.canonicaldict, self.doublecheck):
                name = "{0} {1}".format(line[0], line[1])
                copies[name] += 1
                if copies[name] == 1:
                    self.n_transcripts[line[0]] += 1
            self.copies = {key: 1 for key, value in copies.items() if value > 1}

        missing = [key for key, value in needed.items() if value]
        if missing:
            missing = ["{0} {1}".format(doublecheck[trans], trans) if trans in doublecheck else trans for trans in missing]
            print("WARNING - {0} not found in reference file".format(", ".join(missing)))        


    def read(self, what):
        copies2 = copy.copy(self.copies)
        with FileContext(self.path, "rb") as f:
            for line in refflatreader(f, self.needed, self.canonicaldict, self.doublecheck):
                transcript = "{0} {1}".format(line[0], line[1]) 
                if transcript in copies2:
                    suffix = " copy {}".format(copies2[transcript])
                    copies2[transcript] += 1
                else:
                    suffix = ""
                name = line[0] if self.n_transcripts[line[0]] == 1 else transcript
                if suffix:
                    name = "{}{}".format(name, suffix)

                chrom = STR_2_CHR[line[2]]          
                strand = MINUS if line[3]=="-" else PLUS
                if what == RefFlat.TRANSCRIPTS:
                    yield TranscriptEntry(chrom, int(line[4])+1-self.splice_site_buffer, int(line[5])+self.splice_site_buffer,  name, strand, transcript)

                elif what == RefFlat.EXONS:
                    exon_numbers = xrange(1,int(line[8])+1) if (line[3] == "+") else xrange(int(line[8]),0,-1)            
                    for start, stop, exon in izip(line[9].rstrip(",").split(","), line[10].rstrip(",").split(","), exon_numbers):
                        yield ExonEntry(chrom, int(start)+1-self.splice_site_buffer, int(stop)+self.splice_site_buffer, name, strand, exon)


class Variants(object):
    DISEASE = 0
    GENE = 1
    MUTATION = 2

    def __init__(self, path, genes=(), disease_names=()):
        self.path = path
        self.genes = set([gene.strip().split()[0] for gene in genes])
        self.disease_dict = {}
        for line in csv.reader(disease_names, delimiter="="):
            if len(line) == 2:
                self.disease_dict[line[0].strip("\" ")] = line[1].strip("\" ")


    def read(self, what):
        if what == Variants.MUTATION:
            EntryClass = MutationEntry
            mutation_disease = defaultdict(set)
        else:
            EntryClass = VariantEntry
            extra = {}
        mutations = {}
        mutation_count = Counter()
        with FileContext(self.path, "rb") as f:
            for line in csv.reader(f, delimiter="\t"):
                try:
                    chrom =STR_2_CHR[line[4]]
                except (IndexError, KeyError):
                    continue
                if self.genes and line[3] not in self.genes:
                    continue

                mutation = "{0} {1} {2}:{3}-{4}".format(line[3], line[8], CHR_2_STR[chrom], line[5], line[6])# gene mutation location
                disease = line[1].strip("\" ")
            
                try:
                    disease = self.disease_dict[disease]
                    if disease == "":
                        continue
                except KeyError:
                    pass

                if what == Variants.MUTATION:
                    name = mutation
                    mutation_disease[mutation].add(disease)
                    extra = {"diseases": disease}
                elif what == Variants.DISEASE:
                    name = disease 
                    mutation = (mutation, name)
                elif what == Variants.GENE:
                    name = line[3]

                mutations[mutation] = EntryClass(chrom, int(line[5]), int(line[6]), name, MINUS if line[7]=="-" else PLUS, 1, **extra)
                mutation_count[mutation] += 1

        for mutation, entry in mutations.items():
            if mutation_count[mutation] > 1:
                replace = {"weight": mutation_count[mutation]}
                if what == Variants.MUTATION:
                    replace["diseases"] = "; ".join(sorted(mutation_disease[mutation]))
                entry = entry._replace(**replace)
            yield(entry)


def allbins(bin, level=0, lower=0):
    higher = lower+(8**level)-1
    if bin <= higher:
        return [bin]
    bins = allbins(bin, level+1, higher+1)
    return bins + [lower + (bins[-1]-higher-1)//8]


def reg2bin(beg, end): # accepts 1 based start and stop
    beg -= 1
    end -= 1
    if ((beg>>14) == (end>>14)):
        return ((1<<15)-1)/7 + (beg>>14)
    if ((beg>>17) == (end>>17)):
        return ((1<<12)-1)/7 + (beg>>17)
    if ((beg>>20) == (end>>20)):
        return ((1<<9 )-1)/7 + (beg>>20)
    if ((beg>>23) == (end>>23)):
        return ((1<<6 )-1)/7 + (beg>>23)
    if ((beg>>26) == (end>>26)):
        return ((1<<3 )-1)/7 + (beg>>26)
    return 0


class Bai(object):

    def __init__(self, fn, chr_2_ref):
        try:
            self.chr_2_ref = chr_2_ref
            with open(fn, "rb") as bai:
                if bai.read(4) != "BAI\1":
                    raise CoverMiException("{} is not a BAI file!".format(os.path.basename(fn)))
                
                n_ref = struct.unpack("<i", bai.read(4))[0] # number of reference sequences
                self.bins = [defaultdict(tuple) for x in xrange(0, n_ref)]
                self.intervals = [() for x in xrange(0, n_ref)]
                for ref_id in xrange(0, n_ref):

                    n_bin = struct.unpack("<i", bai.read(4))[0] # number of bins for that reference sequence
                    for distinct_bin in xrange(0, n_bin):
                    
                        bin, n_chunk = struct.unpack("<Ii", bai.read(8)) # bin_number, number of chunks within that bin
                        if n_chunk:
                            if bin == 37450: # pseudo bin for unmapped reads
                                bai.read(32)
                            else:
                                chunks = [None] * n_chunk
                                for chunk in xrange(0, n_chunk):
                                    chunks[chunk] = struct.unpack("<QQ", bai.read(16))
                                self.bins[ref_id][bin] = chunks

                    n_intv = struct.unpack("<i", bai.read(4))[0] # length of linear index
                    if n_intv:
                        intervals = [None] * n_intv
                        for intv in xrange(0, n_intv):
                            intervals[intv] = struct.unpack("<Q", bai.read(8))[0]
                        self.intervals[ref_id] = intervals

        except IOError:
            raise CoverMiException("IOError reading {}".format(os.path.basename(fn)))

        except struct.error:
            raise CoverMiException("{} is truncated".format(os.path.basename(fn)))
        


    def chunks(self, locations):
        chunks = set()
        for loc in locations:
            ref_id = self.chr_2_ref[loc.chrom]
            lower_bound = self.intervals[ref_id][(loc.start-1)//16384]
            for bin in allbins(reg2bin(loc.start, loc.stop)):
                for vstartstop in self.bins[ref_id][bin]:
                    if vstartstop[1] >= lower_bound:
                        chunks.add(vstartstop)
        return sorted(chunks)


BamStats = namedtuple("BamStats", "mapq unmapped duplicate")


class Bam(object):

    COVERAGE = 1
    STATS = 2

    Q30 = 1
    READ_LENGTH = 2

    def __init__(self, fn, index=True):
        self.bam = None
        self.bai = None

        try:
            self.bam = Bio.bgzf.open(fn, "rb") if BGZF else BufferedReader(gzip.open(fn, "rb"))
            if self.bam.read(4) != "BAM\1":
                self.bam.close()
                raise CoverMiException("{} is not a BAM file!".format(os.path.basename(fn)))
            
            len_header_text = struct.unpack("<i", self.bam.read(4))[0]
            header_text = self.bam.read(len_header_text)
            num_ref_seq = struct.unpack("<i", self.bam.read(4))[0]
            chr_2_ref = [None] * num_ref_seq
            self.ref_2_chr = [None] * num_ref_seq
            for x in xrange(0, num_ref_seq):
                len_ref_name = struct.unpack("<i", self.bam.read(4))[0]
                ref_name = self.bam.read(len_ref_name - 1)
                chrom = STR_2_CHR[ref_name]
                self.ref_2_chr[x] = chrom
                chr_2_ref[chrom] = x
                self.bam.read(5)

        except IOError:
            if self.bam: self.bam.close()
            raise CoverMiException("IOError reading {}".format(os.path.basename(fn)))

        except (IOError, struct.error):
            if self.bam: self.bam.close()
            raise CoverMiException("{} has a truncated header".format(os.path.basename(fn)))
             
        if index:
            if BGZF:
                try:
                    self.bai = Bai(fn+".bai", chr_2_ref)
                    print("Using index file {}".format(os.path.basename(fn+".bai")))
                except CoverMiException as e:
                    print(str(e))
            else:
                print("Biopython not installed therefore cannot use bam indexes")

    def __enter__(self):
        return self


    def __exit__(self, type, value, traceback):
        self.bam.close()


    def coverage(self, locations=(), print_progress=False):
        self.unmapped = 0
        self.mapped = 0
        if locations and self.bai:
            chunks = self.bai.chunks(locations)
        else:
            chunks = ((0, 0),)
            print_progress = False
        try:
            for chunk_no, chunk in enumerate(chunks):
                vstart, vstop = chunk
                if vstart:
                    self.bam.seek(vstart)
                if print_progress:
                    sys.stdout.write("Reading chunk {} of {}\r".format(chunk_no+1, len(chunks)))
                    sys.stdout.flush()
                while True:
                    if vstop and self.bam.tell() > vstop:
                        break
                    read = self.bam.read(36)
                    if len(read) == 0:
                        break
                    
                    block_size, ref_id, pos, bin_mq_nl, flag_nc, len_seq, next_ref_id, next_pos, len_template = struct.unpack("<iiiIIiiii", read)
                    flag = flag_nc >> 16#// 0x10000
                    unmapped = (ref_id == -1) or (flag & 0x4)
#                    duplicate = flag & 0x400
#                    secondary = flag & 0x100
#                    supplementary = flag & 0x800

                    if unmapped: # unmapped read
                        self.bam.read(block_size-32)
                        self.unmapped += 1
                    else:
                        self.mapped += 1
                        len_read_name = bin_mq_nl & 0xFF
                        n_cigar_op = flag_nc & 0xFFFF
                        direction = MINUS if flag & 0x10 else PLUS
                        start = pos + 1

                        read_name = self.bam.read(len_read_name - 1)
                        self.bam.read(1)

                        cigar_bytes = n_cigar_op * 4
                        length = 0
                        for cigar in struct.unpack("<" + "I" * n_cigar_op, self.bam.read(cigar_bytes)):
                            cigar_op = cigar & 0xF
                            if cigar_op in (0, 2, 7, 8):
                                length += cigar // 0x10
                            elif cigar_op == 3: # skip an intron
                                if length:
                                    yield (self.ref_2_chr[ref_id], start, start + length - 1, direction)
                                start += length + (cigar//0x10)
                                length = 0
                        if length:
                            yield (self.ref_2_chr[ref_id], start, start + length - 1, direction)

                        self.bam.read(block_size - 32 - len_read_name - cigar_bytes)
            if print_progress:
                print()

        except (IOError, struct.error):
            self.bam.close()
            raise CoverMiException("{} is truncated".format(os.path.basename(fn)))


    def read(self, reference=None, locations=()):
        try:
            for vstart, vstop in self.bai.chunks(locations) if locations and self.bai else ((0, 0),):
                if vstart:
                    self.bam.seek(vstart)
                while True:
                    if vstop and bam.tell() > vstop:
                        break
                    read = self.bam.read(36)
                    if len(read) == 0:
                        break
                    
                    block_size, ref_id, pos, bin_mq_nl, flag_nc, len_seq, next_ref_id, next_pos, len_template = struct.unpack("<iiiIIiiii", read)

                    flag = flag_nc // 0x10000
                    unmapped = (ref_id == -1) or flag & 0x10
                    duplicate = flag & 0x400
                    secondary = flag & 0x100
                    supplementary = flag & 0x800

                    len_read_name = 0
                    cigar_bytes = 0

                    if what ==  Bam.STATS:
                        mapq = (bin_mq_nl // 0x100) & 0xFF
                        yield BamStats(mapq, unmapped, duplicate)

                    elif what == Bam.COVERAGE and not (unmapped or duplicate or secondary or supplementary):
                        len_read_name = bin_mq_nl & 0xFF
                        n_cigar_op = flag_nc & 0xFFFF

                        direction = MINUS if flag & 0x10 else PLUS
                        start = pos + 1

                        read_name = self.bam.read(len_read_name - 1)
                        self.bam.read(1)

                        cigar_bytes = n_cigar_op * 4
                        length = 0
                        for cigar in struct.unpack("<" + "I" * n_cigar_op, self.bam.read(cigar_bytes)):
                            cigar_op = cigar & 0xF
                            if cigar_op in (0, 2, 7, 8):
                                length += cigar // 0x10
                            elif cigar_op == 3: # skip an intron
                                if length:
                                    yield (self.ref_2_chr[ref_id], start, start + length - 1, direction)
                                start += length + (cigar//0x10)
                                length = 0
                        if length:
                            yield (self.ref_2_chr[ref_id], start, start + length - 1, direction)

                    self.bam.read(block_size - 32 - len_read_name - cigar_bytes)

        except (IOError, struct.error):
            self.bam.close()
            raise CoverMiException("{} is truncated".format(fn))


    def info(self, what, locations=()):
        if what not in (Bam.Q30, Bam.READ_LENGTH):
            raise CoverMiException("Bam: Unknown option: {}".format(what))
        passing_q30 = 0
        total = 0
        try:
            for vstart, vstop in self.bai.chunks(locations) if locations and self.bai else ((0, 0),):
                if vstart:
                    self.bam.seek(vstart)
                while True:
                    if vstop and bam.tell() > vstop:
                        break
                    read = self.bam.read(36)
                    if len(read) == 0:
                        break
                    
                    block_size, ref_id, pos, bin_mq_nl, flag_nc, len_seq, next_ref_id, next_pos, len_template = struct.unpack("<iiiIIiiii", read)

                    flag = flag_nc // 0x10000
                    unmapped = (ref_id == -1) or flag & 0x10
                    duplicate = flag & 0x400
                    secondary = flag & 0x100
                    supplementary = flag & 0x800
                    n_cigar_op = flag_nc & 0xFFFF

                    if unmapped or duplicate or secondary or supplementary or n_cigar_op==0:
                        self.bam.read(block_size-32)
                        continue

                    len_read_name = bin_mq_nl & 0xFF
                    self.bam.read(len_read_name)

                    cigar_bytes = n_cigar_op * 4
                    cigar_string = [(cigar & 0xF, cigar // 0x10) for cigar in struct.unpack("<" + "I" * n_cigar_op, self.bam.read(cigar_bytes))]

                    start = 1 if cigar_string[0][0] == 5 else 0
                    if cigar_string[start][0] == 4:
                        soft_clipped = cigar_string[start][1]
                        start += 1
                    else:
                        soft_clipped = 0
                    length = sum([cigar[1] for cigar in cigar_string[start:] if cigar[0] in (0, 2, 7, 8)])
                    if what == Bam.READ_LENGTH: return length

                    self.bam.read((len_seq+1)/2)
                    quality = struct.unpack("<" + "B" * len_seq, self.bam.read(len_seq))[soft_clipped:length+soft_clipped]
                    for qual in quality:
                        if qual >= 30: passing_q30 += 1
                    total += len(quality)

                    self.bam.read(block_size - 32 - len_read_name - cigar_bytes - ((len_seq+1)/2) - len_seq)

        except (IOError, struct.error):
            self.bam.close()
            raise CoverMiException("{} is truncated".format(fn))

        return passing_q30 * 100.0 / total if total>0 else 0
